/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a SANE backend to access the SANE network daemon.
 */

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>
#include <netsane.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <sane.h>
#include <sanei.h>

DBG_DECL (METANET)

#define PATH_NETSANE_CONFIG PATH_SANE_CONFIG_DIR "/metanet.conf"

static char *defaulthost = "localhost";

static int peer = -1;		/* XXX fixme! */
static int opt_reload = 1;

/* checked. 100% o.k. */
static int
openit (char *host)
{
  struct in_addr defaddr;
  struct hostent *hp, def;
  struct servent *sp;
  struct sockaddr_in sin;
  int s;
  char *alist[1];

  if (!(hp = gethostbyname (host)))
    {
      defaddr.s_addr = inet_addr (host);
      if (!inet_aton (host, &defaddr))
	{
	  DBG (METANET, 1, "metanet.openit: unknown host: %s\n", host);
	  return 0;
	}
      def.h_name = host;
      def.h_addr_list = alist;
      def.h_addr = (char *) &defaddr;
      def.h_length = sizeof (struct in_addr);
      def.h_addrtype = AF_INET;
      def.h_aliases = 0;
      hp = &def;
    }
  if (!(sp = getservbyname ("netsane", "tcp")))
    {
      DBG (METANET, 1, "metanet.openit: tcp/netsane: unknown service\n");
      return 0;
    }
  sin.sin_family = hp->h_addrtype;
  memcpy (&sin.sin_addr, hp->h_addr, hp->h_length);
  sin.sin_port = sp->s_port;
  if ((s = socket (hp->h_addrtype, SOCK_STREAM, 0)) < 0)
    {
      DBG (METANET, 1, "metanet.openit: socket: %s", strerror (errno));
      return 0;
    }

  /* have network connection; identify the host connected with */
  DBG (METANET, 1, "metanet.openit: connecting to %s\n", hp->h_name);
  if (connect (s, (struct sockaddr *) &sin, sizeof (sin)) < 0)
    {
      DBG (METANET, 1, "metanet.openit: connect: %s", strerror (errno));
      close (s);
      return 0;
    }
  return s;
}

#ifndef NDEBUG

/* checked. debug only. 100% o.k. */
static void
chkcmd (int a, int b)
{
  if (a != b)
    fprintf (stderr, "Command mismatch %x != %x .\n", a, b);
}

#endif /* !NDEBUG */

/* checked. must be called after openit. */
static void
say_helo (void)
{
  int cmd;
  char buffer[1024];

  {
    char hostname[128];
    if (gethostname (hostname, sizeof (hostname)))
      *hostname = '\0';
    sprintf (buffer, "Hi ! This is %s calling.", hostname);
    cmd = CMD_HELO;
    sanei_printf (peer, "%d %s\n", &cmd, buffer);
    sanei_scanf (peer, "%d %s\n", &cmd, buffer);
    IF_DBG (chkcmd (cmd, CMD_HELO));
    DBG (METANET, 1, "metanet.say_helo: remote HELO is `%s'.\n", buffer);
  }
}

/* checked. 100% o.k. */
static void
say_goodbye (void)
{
  int cmd;
  cmd = CMD_QUIT;
  sanei_printf (peer, "%d\n", &cmd);
}

static char *
readconfig (char *filename)
{
  FILE *fp;
  static char buffer[PATH_MAX];
  size_t len;

  DBG_INIT (METANET);

  fp = fopen (filename, "r");

  if (fp)
    {
      while (fgets (buffer, sizeof (buffer), fp))
	{
	  if (buffer[0] == '#')	/* ignore line comments */
	    continue;
	  len = strlen (buffer);
	  if (buffer[len - 1] == '\n')
	    buffer[--len] = '\0';
	  if (len == 0)
	    continue;		/* ignore empty lines */
	  return buffer;
	}
      fclose (fp);
    }
  return NULL;
}

/* checked. */
SANE_Status
sane_init (SANE_Int * version_code, SANE_Authorization_Callback sac)
{
  int cmd;
  SANE_Status bla;
  SANE_Int vc2;

  char dev_name[PATH_MAX];
  char *host;

  DBG_INIT (METANET);

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);
  /* Just in case we cannot connect */

  /* default to defaulthost instead of insisting on config file */
  strcpy (dev_name, defaulthost);

  host = readconfig (PATH_NETSANE_CONFIG);
  if (host)
    strcpy (dev_name, host);

  host = getenv ("METANETHOST");
  if (host)
    strcpy (dev_name, host);

  peer = openit (dev_name);
  if (peer < 0)
    return SANE_STATUS_IO_ERROR;	/* Connection failed. */

  say_helo ();

  cmd = CMD_SANE_INIT;
  sanei_printf (peer, "%d\n", &cmd);
  sanei_scanf (peer, "%d %d %d\n", &cmd, &bla, &vc2);
  IF_DBG (chkcmd (cmd, CMD_SANE_INIT));
  if (version_code)
    *version_code = vc2;

  return bla;
}

/* checked. 100% o.k. */
void
sane_exit (void)
{
  int cmd;

  DBG (METANET, 1, "metanet.exit()\n");
  cmd = CMD_SANE_EXIT;
  sanei_printf (peer, "%d\n", &cmd);
  sanei_scanf (peer, "%d\n", &cmd);
  IF_DBG (chkcmd (cmd, CMD_SANE_EXIT));
  say_goodbye ();
}

#if 0				/* Enable this for remote strstatus resolution */
/* Do we need to cache these, too ? */
const SANE_String_Const
sane_strstatus (SANE_Status status)
{
  int cmd;
  static char buffer[1024];

  cmd = CMD_SANE_STRS;
  ns_printf ("%d %d\n", &cmd, &status);
  ns_scanf ("%d %s\n", &cmd, buffer);
  debug (chkcmd (cmd, CMD_SANE_STRS));

  return buffer;
}
#endif

SANE_Status
sane_get_devices (const SANE_Device *** device_list, SANE_Bool local_only)
{
  int cmd, x;
  SANE_Status bla;
  static SANE_Device devs[256];
  static const SANE_Device *devlist[256];

  DBG (METANET, 2, "metanet.get_devices: local_only=%d\n", local_only);

  if (local_only)
    {
      *device_list = 0;
      return SANE_STATUS_GOOD;
    }

  cmd = CMD_SANE_GDEV;
  sanei_printf (peer, "%d\n", &cmd);
  sanei_scanf (peer, "%d %d", &cmd, &bla);
  IF_DBG (chkcmd (cmd, CMD_SANE_GDEV));

  x = 0;
  while (1)
    {
      if (sanei_scanf (peer, "%&s", &devs[x].name) == 0)
	break;
      sanei_scanf (peer, "%&s", &devs[x].vendor);
      sanei_scanf (peer, "%&s", &devs[x].model);
      sanei_scanf (peer, "%&s", &devs[x].type);
      devlist[x] = devs + x;
      x++;
    }
  devlist[x] = NULL;
  sanei_scanf (peer, "\n");

  *device_list = devlist;
  return bla;
}

/* checked. 100% o.k. */
SANE_Status
sane_open (SANE_String_Const devicename,
	   SANE_Handle * handle)
{
  int cmd;
  SANE_Status bla;

  opt_reload = 1;
  cmd = CMD_SANE_OPEN;
  sanei_printf (peer, "%d %s\n", &cmd, devicename);
  sanei_scanf (peer, "%d %d %*c\n",
	       &cmd, &bla, sizeof (SANE_Handle), handle);
  IF_DBG (chkcmd (cmd, CMD_SANE_OPEN));

  return bla;
}

/* checked. 100% o.k. */
void
sane_close (SANE_Handle handle)
{
  int cmd;

  opt_reload = 1;
  cmd = CMD_SANE_CLOS;
  sanei_printf (peer, "%d %*c\n", &cmd, sizeof (handle), &handle);
  sanei_scanf (peer, "%d\n", &cmd);
  IF_DBG (chkcmd (cmd, CMD_SANE_CLOS));
}

static void
free_opt (SANE_Option_Descriptor * od)
{
  free ((void *) od->name);
  free ((void *) od->title);
  free ((void *) od->desc);
  if (od->constraint.word_list)
    free ((void *) od->constraint.word_list);
  free (od);
}

static void
copy_opt (SANE_Option_Descriptor ** new, SANE_Option_Descriptor * old)
{
  int constr_length;

  if ((*new = malloc (sizeof (SANE_Option_Descriptor))) == NULL)
    return;

  **new = *old;			/* Copy the static parts */

  (*new)->name = strdup (old->name);
  (*new)->title = strdup (old->title);
  (*new)->desc = strdup (old->desc);

  constr_length = 0;
  switch (old->constraint_type)
    {
    case SANE_CONSTRAINT_NONE:
      break;

    case SANE_CONSTRAINT_STRING_LIST:
      for (constr_length = 0;
	   old->constraint.string_list[constr_length];
	   constr_length++);
      constr_length++;
      constr_length *= sizeof (SANE_String_Const);
      break;

    case SANE_CONSTRAINT_WORD_LIST:
      constr_length = (old->constraint.word_list[0] + 1) * sizeof (SANE_Int);
      break;

    case SANE_CONSTRAINT_RANGE:
      constr_length = 3 * sizeof (SANE_Int);
      break;
    }
  if (constr_length == 0)
    (*new)->constraint = NULL;
  else
    {
      if (((*new)->constraint.word_list = malloc (constr_length)) == NULL)
	return;			/* fixme */
      memcpy ((void *) (*new)->constraint.word_list, old->constraint.word_list,
	      constr_length);
    }
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  int cmd;
  static SANE_Option_Descriptor *opts[256];
  static SANE_Option_Descriptor od;
  static char name[1024], title[1024], desc[1024];
  static int ranges[256];
  static SANE_String_Const strrange[256];
  static SANE_Range range;
  static SANE_Handle lasthand;

  /* Caching */
  if (opt_reload || handle != lasthand)
    {
      /* Reload requested or handle changed? */
      memset (opts, 0, sizeof (opts));
      opt_reload = 0;
      lasthand = handle;
    }

  if (opts[option])
    return opts[option];

  /* Option not cached. Get it. */
  cmd = CMD_SANE_GOPT;
  sanei_printf (peer, "%d %*c %d\n", &cmd,
		sizeof (handle), &handle, &option);
  sanei_scanf (peer, "%d", &cmd);
  IF_DBG (chkcmd (cmd, CMD_SANE_GOPT));

  if (0 == sanei_scanf (peer, "%s", name))
    {
      sanei_scanf (peer, "\n");
      return (NULL);
    }
  od.name = name;
  sanei_scanf (peer, "%s", title);
  od.title = title;
  sanei_scanf (peer, "%s", desc);
  od.desc = desc;

  sanei_scanf (peer, "%d", &od.type);
  sanei_scanf (peer, "%d", &od.unit);
  sanei_scanf (peer, "%d", &od.size);
  sanei_scanf (peer, "%d", &od.cap);
  sanei_scanf (peer, "%d", &od.constraint_type);

  switch (od.constraint_type)
    {
    case SANE_CONSTRAINT_NONE:
      break;

    case SANE_CONSTRAINT_STRING_LIST:
      cmd = 0;
      while (sanei_scanf (peer, "%&s", &strrange[cmd++]));
      strrange[cmd++] = NULL;
      od.constraint.string_list = strrange;
      break;

    case SANE_CONSTRAINT_WORD_LIST:
      cmd = 0;
      while (sanei_scanf (peer, "%d", &ranges[cmd++]));
      od.constraint.word_list = ranges;
      break;

    case SANE_CONSTRAINT_RANGE:
      sanei_scanf (peer, "%d %d %d", &range.min, &range.max, &range.quant);
      od.constraint.range = &range;
      break;
    }
  sanei_scanf (peer, "\n");

  if (opts[option])
    free_opt (opts[option]);
  copy_opt (&opts[option], &od);

  return opts[option];
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option,
		     SANE_Action action,
		     void *value,
		     SANE_Int * info)
{
  int cmd;
  SANE_Status bla;
  SANE_Int info2;
  SANE_Value_Type type;
  const SANE_Option_Descriptor *od;

  cmd = CMD_SANE_COPT;

  od = sane_get_option_descriptor (handle, option);	/* Find out the type */
  type = od->type;

  sanei_printf (peer, "%d %*c %d %d %d",
		&cmd, sizeof (handle), &handle, &option, &action, &type);
  switch (action)
    {
    case SANE_ACTION_SET_VALUE:
      switch (type)
	{
	case SANE_TYPE_BOOL:
	case SANE_TYPE_INT:
	case SANE_TYPE_FIXED:
	  sanei_printf (peer, "%d", value);
	  break;

	case SANE_TYPE_STRING:
	case SANE_TYPE_GROUP:
	  sanei_printf (peer, "%s", value);
	  break;

	default:
	  DBG (METANET, 1, "metanet.control_option: unknown type %d\n", type);
	  break;
	}
      break;

    default:
      break;
    }
  sanei_printf (peer, "\n");

  sanei_scanf (peer, "%d %d %d %d", &cmd, &bla, &info2, &type);
  IF_DBG (chkcmd (cmd, CMD_SANE_COPT));
  if (info2 & SANE_INFO_RELOAD_OPTIONS)
    opt_reload = 1;
  if (info)
    *info = info2;

  switch (type)
    {
    case SANE_TYPE_BUTTON:
      break;

    case SANE_TYPE_BOOL:
    case SANE_TYPE_INT:
    case SANE_TYPE_FIXED:
      sanei_scanf (peer, "%d", value);
      break;

    case SANE_TYPE_STRING:
    case SANE_TYPE_GROUP:
      sanei_scanf (peer, "%s", value);
      break;
    }
  sanei_scanf (peer, "\n");

  return bla;
}


/* checked. 100% o.k. */
SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
  int cmd;
  SANE_Status bla;

  cmd = CMD_SANE_PARM;
  sanei_printf (peer, "%d %*c\n", &cmd, sizeof (handle), &handle);
  sanei_scanf (peer, "%d %d %d %d %d %d %d %d\n", &cmd, &bla,
	       &params->format, &params->last_frame,
	       &params->bytes_per_line, &params->pixels_per_line,
	       &params->lines, &params->depth);
  IF_DBG (chkcmd (cmd, CMD_SANE_PARM));

  return bla;
}

/* checked. 100% o.k. */
SANE_Status
sane_start (SANE_Handle handle)
{
  int cmd;
  SANE_Status bla;

  cmd = CMD_SANE_STAR;
  sanei_printf (peer, "%d %*c\n", &cmd, sizeof (SANE_Handle), &handle);
  sanei_scanf (peer, "%d %d\n", &cmd, &bla);
  IF_DBG (chkcmd (cmd, CMD_SANE_STAR));

  return bla;
}

/* checked. 100% o.k. */
SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data,
	   SANE_Int max_length, SANE_Int * length)
{
  int cmd;
  SANE_Status bla;

  cmd = CMD_SANE_READ;
  sanei_printf (peer, "%d %*c %d\n",
		&cmd, sizeof (SANE_Handle), &handle, &max_length);
  sanei_scanf (peer, "%d %d %d", &cmd, &bla, length);
  IF_DBG (chkcmd (cmd, CMD_SANE_READ));
  sanei_scanf (peer, "%*c", *length, data);
  sanei_scanf (peer, "\n");

  return bla;
}

/* checked. 100% o.k. */
extern void
sane_cancel (SANE_Handle handle)
{
  int cmd;

  cmd = CMD_SANE_ABRT;
  sanei_printf (peer, "%d %*c\n", &cmd, sizeof (SANE_Handle), &handle);
}

/* checked. 100% o.k. */
extern SANE_Status
sane_set_io_mode (SANE_Handle handle,
		  SANE_Bool non_blocking)
{
  return SANE_STATUS_UNSUPPORTED;
}

extern SANE_Status
sane_get_select_fd (SANE_Handle handle,
		    SANE_Int * fd)
{
  if (peer < 0)
    return SANE_STATUS_INVAL;

  *fd = peer;
  return SANE_STATUS_GOOD;
}
