/* gtk/SANE-glue -- gtk interfacing routines for SANE
   Uses the SANE library.
   Copyright (C) 1997 David Mosberger and Tristan Tarrant

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <assert.h>
#include <gtk.h>
#include <gtkglue.h>
#include <memory.h>
#include <sane.h>
#include <stdlib.h>
#include <stdio.h>

static int showing_error = 0;

/* forward declarations: */
static void panel_rebuild (GSGDialog *dialog);


static void
set_option (GSGDialog *dialog, int opt_num, void *val, SANE_Action action)
{
  SANE_Status status;
  SANE_Int info;
  char buf[256];

  status = sane_control_option (dialog->dev, opt_num, action, val, &info);
  if (status != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Failed to set value of option %s: %s.",
	       sane_get_option_descriptor(dialog->dev, opt_num)->name,
	       sane_strstatus (status));
      gsg_error (buf);
    }
  if (info & SANE_INFO_RELOAD_OPTIONS)
    panel_rebuild (dialog);
  if ((info & SANE_INFO_RELOAD_PARAMS) && dialog->param_change_callback)
    (*dialog->param_change_callback)(dialog, dialog->param_change_arg);
}

void
gsg_close_dialog_callback (GtkWidget *widget,
			   gpointer client_data, gpointer call_data)
{
  gtk_widget_destroy (client_data);
  showing_error = 0;
}

void
gsg_error (gchar * error)
{
  GtkWidget *main_vbox, *label;
  GtkWidget *button, *error_dialog;

  if (showing_error)
    {
      fprintf (stderr, "%s\n", error);
      return;
    }
  showing_error = 1;
  error_dialog = gtk_window_new ("Error", GTK_WINDOW_TOPLEVEL);

  /* create the main vbox */
  main_vbox = gtk_vbox_new (TRUE, 5);
  gtk_container_set_border_width (main_vbox, 5);
  gtk_widget_show (main_vbox);

  gtk_container_add (error_dialog, main_vbox);

  /* the message */
  label = gtk_label_new (error);
  gtk_container_add (main_vbox, label);
  gtk_widget_show (label);

  /* the confirmation button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button),
		    gsg_close_dialog_callback, error_dialog);
  label = gtk_label_new ("OK");
  gtk_container_add (button, label);
  gtk_container_add (main_vbox, button);

  gtk_widget_show (label);
  gtk_widget_show (button);
  gtk_widget_show (error_dialog);
}

static gint
autobutton_update (GtkObserver *observer, GtkData *data)
{
  GSGDialogElement *elem = observer->user_data;
  GtkDataInt *int_data = (GtkDataInt *) data;
  GSGDialog *dialog = elem->dialog;
  int opt_num = elem - dialog->element;
  const SANE_Option_Descriptor *opt;
  SANE_Status status;
  SANE_Word val;
  char buf[256];

  opt = sane_get_option_descriptor (dialog->dev, opt_num);
  switch (int_data->value)
    {
    case GTK_STATE_ACTIVATED:
      /* turn on automatic mode; */
      set_option (dialog, opt_num, 0, SANE_ACTION_SET_AUTO);
      break;

    case GTK_STATE_DEACTIVATED:
      /* turn off automatic mode by first obtaining current value, then
	 setting that value: */
      status = sane_control_option (dialog->dev, opt_num,
				    SANE_ACTION_GET_VALUE, &val, 0);
      if (status != SANE_STATUS_GOOD)
	{
	  sprintf (buf, "Failed to obtain value of option %s: %s.",
		   opt->name, sane_strstatus (status));
	  gsg_error (buf);
	}
      set_option (dialog, opt_num, &val, SANE_ACTION_SET_VALUE);
      break;
    }
  return FALSE;
}

static void
autobutton_disconnect (GtkObserver *observer, GtkData *data)
{
  free (observer);
}

static void
autobutton_new (GtkWidget *parent, GSGDialogElement *elem,
		GtkWidget *label)
{
  GtkWidget *button, *alignment;
  GtkObserver *observer;

  button = gtk_check_button_new ();
  gtk_container_set_border_width (button, 0);
  gtk_widget_set_usize (button, 20, 20);

  observer = malloc (sizeof (*observer));
  observer->update = autobutton_update;
  observer->disconnect = autobutton_disconnect;
  observer->user_data = elem;
  gtk_data_attach (gtk_button_get_state (button), observer);

  alignment = gtk_alignment_new (0.0, 1.0, 0.5, 0.5);
  gtk_container_add (alignment, button);

  gtk_box_pack (parent, label, FALSE, FALSE, 0, GTK_PACK_END);
  gtk_box_pack (parent, alignment, FALSE, FALSE, 2, GTK_PACK_END);

  gtk_widget_show (alignment);
  gtk_widget_show (button);
}

static gint
button_update (GtkObserver *observer, GtkData *data)
{
  GSGDialogElement *elem = observer->user_data;
  GtkDataInt *int_data = (GtkDataInt *) data;
  GSGDialog *dialog = elem->dialog;
  int opt_num = elem - dialog->element;
  const SANE_Option_Descriptor *opt;
  SANE_Word val = SANE_FALSE;

  opt = sane_get_option_descriptor (dialog->dev, opt_num);
  switch (int_data->value)
    {
    case GTK_STATE_ACTIVATED:
      val = SANE_TRUE;
      /* fall through */
    case GTK_STATE_DEACTIVATED:
      set_option (dialog, opt_num, &val, SANE_ACTION_SET_VALUE);
      break;
    }
  return FALSE;
}

static void
button_disconnect (GtkObserver *observer, GtkData *data)
{
  free (observer);
}

static void
button_new (GtkWidget *parent, const char *name, GSGDialogElement *elem)
{
  GtkWidget *button, *label;
  GtkObserver *observer;

  button = gtk_check_button_new ();

  observer = malloc (sizeof (*observer));
  observer->update = button_update;
  observer->disconnect = button_disconnect;
  observer->user_data = elem;
  gtk_data_attach (gtk_button_get_state (button), observer);

  label = gtk_label_new ((char *) name);
  gtk_label_set_alignment (label, 0.0, 1.0);
  gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);

  gtk_widget_show (label);
  gtk_widget_show (button);

  elem->widget = button;
  elem->label = label;
}

static gint
scale_update (GtkObserver *observer, GtkData *data)
{
  GtkDataAdjustment *adj_data = (GtkDataAdjustment *) data;
  GSGDialogElement *elem = observer->user_data;
  const SANE_Option_Descriptor *opt;
  GSGDialog *dialog = elem->dialog;
  SANE_Word val, new_val;
  int opt_num;

  opt_num = elem - dialog->element;
  opt = sane_get_option_descriptor (dialog->dev, opt_num);
  switch (opt->type)
    {
    case SANE_TYPE_INT:
      val = adj_data->value + 0.5;
      break;

    case SANE_TYPE_FIXED:
      val = SANE_FIX(adj_data->value);
      break;

    default:
      fprintf (stderr, "scale_update: unknown type %d\n", opt->type);
      return FALSE;
    }
  set_option (dialog, opt_num, &val, SANE_ACTION_SET_VALUE);
  sane_control_option (dialog->dev, opt_num, SANE_ACTION_GET_VALUE, &new_val,
		       0);
  if (new_val != val)
    {
      val = new_val;
      goto value_changed;
    }
  return FALSE;		/* value didn't change */

value_changed:
  switch (opt->type)
    {
    case SANE_TYPE_INT:
      adj_data->value = val;
      break;

    case SANE_TYPE_FIXED:
      adj_data->value = SANE_UNFIX(val);
      break;

    default:
      break;
    }
  return TRUE;	/* value did change */
}

static void
scale_disconnect (GtkObserver *observer, GtkData *data)
{
  free (observer);
}

static void
scale_new (GtkWidget *parent, const char *name, gfloat val,
	   gfloat min, gfloat max, gfloat quant, int automatic,
	   GSGDialogElement * elem)
{
  GtkWidget *hbox, *label, *scale;
  GtkObserver * observer;
  GtkData *scale_data;

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_container_set_border_width (hbox, 0);
  gtk_box_pack (parent, hbox, FALSE, FALSE, 0, GTK_PACK_START);

  label = gtk_label_new ((char *) name);
  gtk_label_set_alignment (label, 0.0, 1.0);
  gtk_box_pack (hbox, label, FALSE, FALSE, 2, GTK_PACK_START);

  scale_data = gtk_data_adjustment_new (val, min, max, quant, 1.0, 1.0);
  scale = gtk_hscale_new ((GtkDataAdjustment *) scale_data);

  if (automatic)
    autobutton_new (hbox, elem, scale);
  else
    gtk_box_pack (hbox, scale, FALSE, FALSE, 0, GTK_PACK_END);

  gtk_scale_set_policy (scale, 1);	/* turn on continuous update */
  gtk_scale_set_value_pos (scale, GTK_POS_TOP);
  if (quant - (int) quant == 0.0)
    gtk_scale_set_digits (scale, 0);
  else
    gtk_scale_set_digits (scale, 1);	/* one place behind decimal point */
  observer = malloc (sizeof (*observer));
  observer->update     = scale_update;
  observer->user_data  = (gpointer) elem;
  observer->disconnect = scale_disconnect;
  gtk_data_attach (scale_data, observer);

  gtk_widget_show (label);
  gtk_widget_show (scale);
  gtk_widget_show (hbox);

  elem->widget = scale;
  elem->label = label;
}

static void
push_button_callback (GtkWidget *widget,
		      gpointer client_data, gpointer call_data)
{
  GSGDialogElement *elem = client_data;
  GSGDialog *dialog = elem->dialog;
  int opt_num;

  opt_num = elem - dialog->element;
  set_option (dialog, opt_num, 0, SANE_ACTION_SET_VALUE);
}

static int
option_menu_lookup (GSGMenuItem menu_items[], const char *string)
{
  int i;

  for (i = 0; strcmp (menu_items[i].label, string) != 0; ++i);
  return i;
}

static void
option_menu_callback (GtkWidget *widget, gpointer client_data,
		      gpointer call_data)
{
  GSGMenuItem *menu_item = client_data;
  GSGDialogElement *elem = menu_item->elem;
  const SANE_Option_Descriptor *opt;
  GSGDialog *dialog = elem->dialog;
  int opt_num;
  double dval;
  SANE_Word val;
  void *valp = &val;

  opt_num = elem - dialog->element;
  opt = sane_get_option_descriptor (dialog->dev, opt_num);
  switch (opt->type)
    {
    case SANE_TYPE_INT:
      sscanf (menu_item->label, "%d", &val);
      break;

    case SANE_TYPE_FIXED:
      sscanf (menu_item->label, "%lg", &dval);
      val = SANE_FIX(dval);
      break;

    case SANE_TYPE_STRING:
      valp = menu_item->label;
      break;

    default:
      fprintf (stderr, "option_menu_callback: unexpected type %d\n",
	       opt->type);
      break;
    }
  set_option (dialog, opt_num, valp, SANE_ACTION_SET_VALUE);
}

static void
option_menu_new (GtkWidget *parent, const char *name, char *str_list[],
		 const char *val, GSGDialogElement *elem)
{
  GtkWidget *hbox, *label, *option_menu, *menu, *item;
  GSGMenuItem *menu_items;
  int i, num_items;

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_container_set_border_width (hbox, 0);
  gtk_box_pack (parent, hbox, FALSE, FALSE, 0, GTK_PACK_START);

  label = gtk_label_new ((char *) name);
  gtk_box_pack (hbox, label, FALSE, FALSE, 2, GTK_PACK_START);

  for (num_items = 0; str_list[num_items]; ++num_items);
  menu_items = malloc (num_items * sizeof (menu_items[0]));

  menu = gtk_menu_new ();
  for (i = 0; i < num_items; ++i)
    {
      item = gtk_menu_item_new_with_label (str_list[i]);
      gtk_container_add (menu, item);
      gtk_callback_add (gtk_menu_item_get_state (item),
			option_menu_callback, menu_items + i);
      gtk_widget_show (item);

      menu_items[i].label = str_list[i];
      menu_items[i].elem = elem;
      menu_items[i].index = i;
    }

  option_menu = gtk_option_menu_new ();
  gtk_box_pack (hbox, option_menu, FALSE, FALSE, 2, GTK_PACK_END);
  gtk_option_menu_set_menu (option_menu, menu);
  gtk_option_menu_set_history (option_menu,
			       option_menu_lookup (menu_items, val));
  gtk_widget_show (label);
  gtk_widget_show (option_menu);
  gtk_widget_show (hbox);

  elem->widget = option_menu;
  elem->label  = label;
  elem->menu_size = num_items;
  elem->menu   = menu_items;
}

static gint
text_entry_callback (guint keyval, guint state, gpointer client_data)
{
  GSGDialogElement *elem = client_data;
  const SANE_Option_Descriptor *opt;
  GSGDialog *dialog = elem->dialog;
  gchar *text;
  int opt_num;
  char *buf;
    
  opt_num = elem - dialog->element;
  opt = sane_get_option_descriptor (dialog->dev, opt_num);

  buf = malloc (opt->size);
  buf[0] = '\0';

  text = gtk_text_entry_get_text (elem->widget);
  if (text)
    strcpy (buf, text);

  set_option (dialog, opt_num, buf, SANE_ACTION_SET_VALUE);
  gtk_text_entry_set_text (elem->widget, buf);

  free (buf);

  if (keyval == XK_Return)
    return TRUE;

  return FALSE;
}

static void
text_entry_new (GtkWidget *parent, const char *name, const char *val,
		GSGDialogElement *elem)
{
  GtkWidget *hbox, *text, *label;

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_container_set_border_width (hbox, 0);
  gtk_box_pack (parent, hbox, FALSE, FALSE, 0, GTK_PACK_START);

  label = gtk_label_new ((char *) name);
  gtk_box_pack (hbox, label, FALSE, FALSE, 2, GTK_PACK_START);

  text = gtk_text_entry_new ();
  gtk_text_entry_set_text (text, (char *) val);
  gtk_text_entry_set_key_function (text, text_entry_callback, elem);
  gtk_box_pack (hbox, text, FALSE, TRUE, 0, GTK_PACK_START);

  gtk_widget_show (hbox);
  gtk_widget_show (label);
  gtk_widget_show (text);

  elem->widget = text;
  elem->label  = label;
}

static void
panel_destroy (GSGDialog *dialog)
{
  const SANE_Option_Descriptor *opt;
  GSGDialogElement *elem;
  int i, j;

  gtk_widget_destroy (dialog->vbox);

  /* free the menu labels of integer/fix-point word-lists: */
  for (i = 0; i < dialog->num_elements; ++i)
    {
      if (dialog->element[i].menu)
	{
	  opt = sane_get_option_descriptor (dialog->dev, i);
	  elem = dialog->element + i;
	  if (opt->type != SANE_TYPE_STRING)
	    for (j = 0; j < elem->menu_size; ++j)
	      if (elem->menu[j].label)
		free (elem->menu[j].label);
	  free (elem->menu);
	}      
    }
}

static void
panel_build (GSGDialog *dialog)
{
  GtkWidget *frame, *parent, *vbox, *button, *label;
  const SANE_Option_Descriptor *opt;
  SANE_Handle dev = dialog->dev;
  GSGDialogElement *elem;
  SANE_Word quant, val;
  SANE_Status status;
  SANE_Int num_words;
  char *buf, str[16];
  char **str_list;
  int i, j;

  /* create main vbox for easy destruction: */
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (vbox, 2);
  gtk_container_add (dialog->window, vbox);
  gtk_widget_show (vbox);
  dialog->vbox = vbox;

  parent = vbox;
  for (i = 1; i < dialog->num_elements; ++i)
    {
      opt = sane_get_option_descriptor (dev, i);
      elem = dialog->element + i;
      elem->dialog = dialog;

      /* adjust visibility: */
      if (!SANE_OPTION_IS_ACTIVE(opt->cap))
	continue;

      switch (opt->type)
	{
	case SANE_TYPE_GROUP:
	  /* group a set of options */
	  frame = gtk_frame_new ((char *) opt->title);
	  gtk_container_set_border_width (frame, 4);
	  gtk_frame_set_type (frame, GTK_SHADOW_ETCHED_IN);
	  gtk_box_pack (vbox, frame, FALSE, TRUE, 0, GTK_PACK_START);
	  gtk_widget_show (frame);

	  parent = gtk_vbox_new (FALSE, 4);
	  gtk_container_set_border_width (parent, 2);
	  gtk_container_add (frame, parent);
	  gtk_widget_show (parent);
	  elem->widget = parent;
	  break;

	case SANE_TYPE_BOOL:
	  assert (opt->size == sizeof(SANE_Word));
	  button_new (parent, opt->title, elem);

	  status = sane_control_option (dialog->dev, i, SANE_ACTION_GET_VALUE,
					&val, 0);
	  if (status != SANE_STATUS_GOOD)
	    goto get_value_failed;

	  gtk_check_button_set_state (elem->widget, val);
	  break;

	case SANE_TYPE_INT:
	  if (opt->size != sizeof(SANE_Word))
	    {
	      fprintf (stderr, "Integer vectors not supported yet\n");
	      break;
	    }
	  status = sane_control_option (dialog->dev, i, SANE_ACTION_GET_VALUE,
					&val, 0);
	  if (status != SANE_STATUS_GOOD)
	    goto get_value_failed;

	  switch (opt->constraint_type)
	    {
	    case SANE_CONSTRAINT_RANGE:
	      /* use a scale */
	      quant = opt->constraint.range->quant;
	      if (quant == 0)
		quant = 1;
	      scale_new (parent, opt->title, val,
			 opt->constraint.range->min,
			 opt->constraint.range->max, quant,
			 (opt->cap & SANE_CAP_AUTOMATIC), elem);
	      break;

	    case SANE_CONSTRAINT_WORD_LIST:
	      /* use a "list-selection" widget */
	      num_words = opt->constraint.word_list[0];
	      str_list = malloc ((num_words + 1) * sizeof (str_list[0]));
	      for (j = 0; j < num_words; ++j)
		{
		  sprintf (str, "%d", opt->constraint.word_list[j + 1]);
		  str_list[j] = strdup (str);
		}
	      str_list[j] = 0;
	      sprintf (str, "%d", val);
	      option_menu_new (parent, opt->title, str_list, str, elem);
	      free (str_list);
	      break;

	    default:
	      fprintf (stderr, "panel_build: unknown constraint %d!\n",
		       opt->constraint_type);
	      break;
	    }
	  break;

	case SANE_TYPE_FIXED:
	  if (opt->size != sizeof(SANE_Word))
	    {
	      fprintf (stderr, "Fixedpoint vectors not supported yet\n");
	      break;
	    }
	  status = sane_control_option (dialog->dev, i, SANE_ACTION_GET_VALUE,
					&val, 0);
	  if (status != SANE_STATUS_GOOD)
	    goto get_value_failed;

	  switch (opt->constraint_type)
	    {
	    case SANE_CONSTRAINT_RANGE:
	      /* use a scale */
	      quant = opt->constraint.range->quant;
	      if (quant == 0)
		quant = 1;
	      scale_new (parent, opt->title, SANE_UNFIX(val),
			 SANE_UNFIX(opt->constraint.range->min),
			 SANE_UNFIX(opt->constraint.range->max),
			 SANE_UNFIX(quant), (opt->cap & SANE_CAP_AUTOMATIC),
			 elem);
	      break;

	    case SANE_CONSTRAINT_WORD_LIST:
	      /* use a "list-selection" widget */
	      num_words = opt->constraint.word_list[0];
	      str_list = malloc ((num_words + 1) * sizeof (str_list[0]));
	      for (j = 0; j < num_words; ++j)
		{
		  sprintf (str, "%g",
			   SANE_UNFIX(opt->constraint.word_list[j + 1]));
		  str_list[j] = strdup (str);
		}
	      str_list[j] = 0;
	      sprintf (str, "%g", SANE_UNFIX(val));
	      option_menu_new (parent, opt->title, str_list, str, elem);
	      free (str_list);
	      break;

	    default:
	      fprintf (stderr, "panel_build: unknown constraint %d!\n",
		       opt->constraint_type);
	      break;
	    }
	  break;

	case SANE_TYPE_STRING:
	  buf = malloc (opt->size);
	  status = sane_control_option (dialog->dev, i, SANE_ACTION_GET_VALUE,
					buf, 0);
	  if (status != SANE_STATUS_GOOD)
	    goto get_value_failed;

	  switch (opt->constraint_type)
	    {
	    case SANE_CONSTRAINT_STRING_LIST:
	      /* use a "list-selection" widget */
	      option_menu_new (parent, opt->title,
			       (char **) opt->constraint.string_list, buf,
			       elem);
	      break;

	    case SANE_CONSTRAINT_NONE:
	      text_entry_new (parent, opt->title, buf, elem);
	      break;

	    default:
	      fprintf (stderr, "panel_build: unknown constraint %d!\n",
		       opt->constraint_type);
	      break;
	    }
	  free (buf);
	  break;

	case SANE_TYPE_BUTTON:
	  button = gtk_push_button_new ();
	  gtk_callback_add (gtk_button_get_state (button),
			    push_button_callback, elem);

	  label = gtk_label_new ((char *) opt->title);
	  gtk_container_add (button, label);

	  gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);

	  gtk_widget_show (label);
	  gtk_widget_show (button);

	  elem->widget = button;
	  elem->label  = label;
	  break;

	default:
	  fprintf (stderr, "panel_build: Unknown type %d\n", opt->type);
	  break;
	}
      continue;

    get_value_failed:
      {
	char msg[256];

	sprintf (msg, "Failed to obtain value of option %s: %s.",
		 opt->name, sane_strstatus (status));
	gsg_error (msg);
      }
    }
  return;

}

/* When an setting an option changes the dialog, everything may
   change: the option titles, the activity-status of the option, its
   constraints or what not.  Thus, rather than trying to be clever in
   detecting what exactly changed, we use a brute-force method of
   rebuilding the entire dialog.  */
static void
panel_rebuild (GSGDialog *dialog)
{
  panel_destroy (dialog);
  panel_build (dialog);
}

GSGDialog*
gsg_create_dialog (GtkWidget *window, const char *device_name,
		   GSGParamChangeCallback callback, void *arg)
{
  SANE_Int num_elements;
  GSGDialog * dialog;
  SANE_Status status;
  SANE_Handle dev;
  char buf[256];

  status = sane_open (device_name, &dev);
  if (status != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Error opening %s: %s.",
	       device_name, sane_strstatus (status));
      gsg_error (buf);
      return 0;
    }

  if (sane_control_option (dev, 0, SANE_ACTION_GET_VALUE, &num_elements, 0)
      != SANE_STATUS_GOOD)
    {
      gsg_error ("Error obtaining option count.");
      sane_close (dev);
      return 0;
    }

  dialog = malloc (sizeof (*dialog));
  memset (dialog, 0, sizeof (*dialog));

  dialog->window = window;
  dialog->dev = dev;
  dialog->num_elements = num_elements;
  dialog->param_change_callback = callback;
  dialog->param_change_arg = arg;

  dialog->element = malloc (num_elements * sizeof (dialog->element[0]));
  memset (dialog->element, 0, num_elements * sizeof (dialog->element[0]));

  panel_build (dialog);
  return dialog;
}

void
gsg_destroy_dialog (GSGDialog *dialog)
{
  SANE_Handle dev = dialog->dev;

  panel_destroy (dialog);
  free (dialog->element);
  free (dialog);

  sane_close (dev);
}
