/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements a simple SANE frontend (well it rather is a
   transport layer, but seen from libsane it is a frontend) which acts
   as a NETSANE server. The NETSANE specifications should have come
   with this package.  
   Feel free to enhance this program ! It needs extension especially 
   regarding crypto-support and authentication.
 */

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>

#include <sane.h>
#include <sanei.h>
#include <netsane.h>

#define NELEMS(a)	((int)(sizeof (a) / sizeof (a[0])))

#undef DEBUG_NETSANE

#ifdef DEBUG_NETSANE
#define DEBUG(x) x
#else
#define DEBUG(x)
#endif

static int peer_in, peer_out;

enum STATE
  {
    NOT_CONN,			/* Connection not yet established */
    HELO_RECV,			/* Got the HELO message from the client */
    SANE_NOINIT,		/* SANE has not yet been initialized  */
    SANE_INIT,			/* SANE-Init has been called */
    SANE_OPEN,			/* SANE-device has been opened */
    SANE_STAR,			/* A scan has been started */
    SANE_QUIT			/* Connection-break requested */
  };

enum AUTHSTATE
  {
    NO_AUTH,			/* Authentication not required/done */
    AUTH_REQ,			/* Authentication necessary but not yet started */
    AUTH_USER,			/* Username has been received */
    AUTH_OK			/* Authentication data for user is ok */
  }
auth_state = NO_AUTH;

enum STATE sane_state = NOT_CONN;
char strbuf[32*1024];

void 
netsane_auth_callback (SANE_String_Const domain,
		       SANE_String_Const *uid, SANE_String_Const *passwd)
{
  /* fixme */
}

void
readcmd (void)
{
  int command, hlp;
  SANE_Int si_hlp;
  SANE_Status ss_hlp;
  const char *schr_hlp;

  sanei_scanf (peer_in, "%d", &command);
  alarm (300);			/* We have received a command. Wait another 5 Minutes ... */
  switch (command)
    {
      /* Setup and Misc easy commands */
    case CMD_HELO:
      sanei_scanf (peer_in, "%s\n", strbuf);
      syslog (LOG_DEBUG, "Helo: %s\n", strbuf);
      sane_state = HELO_RECV;
      break;

    case CMD_QUIT:
      syslog (LOG_DEBUG, "QUIT.\n");
      sane_state = SANE_QUIT;
      break;

      /* Authentication */
    case CMD_USER_REQ:
      syslog (LOG_DEBUG,
	      "Error: client asked server to authenticate. Not yet !\n");
      break;

    case CMD_USER_RPLY:
      sanei_scanf (peer_in, "%s\n", strbuf);
      switch (auth_state)
	{
	case NO_AUTH:
	  syslog (LOG_DEBUG, "AUTH not requested but got USER. Ignoring.\n");
	  break;

	case AUTH_OK:
	  syslog (LOG_DEBUG, "Re-authenticating as %s\n", strbuf);
	case AUTH_USER:
	case AUTH_REQ:
	  auth_state = AUTH_USER;
	  /* Fixme real auth here ! */
	  command = CMD_PASS_REQ;
	  hlp = PASS_UNIX;
	  /* request Auth */
	  sanei_printf (peer_out, "%d %d\n", &command, &hlp);
	  break;
	}
      break;

    case CMD_PASS_REQ:
      syslog (LOG_DEBUG,
	      "Error: client asked server to authenticate. Not yet !\n");
      break;

    case CMD_PASS_RPLY:
      sanei_scanf (peer_in, "%s\n", strbuf);
      switch (auth_state)
	{
	case NO_AUTH:
	  syslog (LOG_DEBUG, "AUTH not requested but got PASS. Ignoring.\n");
	  break;

	case AUTH_REQ:
	  syslog (LOG_DEBUG, "USER unset but got PASS. Ignoring.\n");
	  break;

	case AUTH_OK:
	  syslog (LOG_DEBUG, "Re-authenticating as %s\n", strbuf);

	case AUTH_USER:
	  auth_state = AUTH_USER;
	  /* Fixme real auth here ! */
	  command = CMD_PASS_REQ;
	  hlp = PASS_UNIX;
	  /* request Auth */
	  sanei_printf (peer_out, "%d %d\n", &command, &hlp);
	  break;
	}
      break;

    case CMD_SANE_INIT:
      sanei_scanf (peer_in, "\n");
      sane_state = SANE_INIT;
      ss_hlp = sane_init (&si_hlp, netsane_auth_callback);
      syslog (LOG_DEBUG, "sane_init: status %s, version_code=%x\n",
	      sane_strstatus (ss_hlp), si_hlp);
      sanei_printf (peer_out, "%d %d %d\n", &command, &ss_hlp, &si_hlp);
      break;

    case CMD_SANE_EXIT:
      syslog (LOG_DEBUG, "sane_exit.\n");
      sanei_scanf (peer_in, "\n");
      sane_state = SANE_NOINIT;
      sane_exit ();
      sanei_printf (peer_out, "%d\n", &command);
      break;

    case CMD_SANE_GDEV:
      {
	const SANE_Device **device_list;
	sanei_scanf (peer_in, "\n");
	ss_hlp = sane_get_devices (&device_list, SANE_TRUE);
	syslog (LOG_DEBUG, "sane_get_devices: %d.\n", ss_hlp);
	sanei_printf (peer_out, "%d %d", &command, &ss_hlp);
	while (*device_list)
	  {
	    sanei_printf (peer_out, "%s", (*device_list)->name);
	    sanei_printf (peer_out, "%s", (*device_list)->vendor);
	    sanei_printf (peer_out, "%s", (*device_list)->model);
	    sanei_printf (peer_out, "%s", (*device_list)->type);
	    device_list++;
	  }
	sanei_printf (peer_out, "\n");
      }
      break;

    case CMD_SANE_OPEN:
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%s\n", strbuf);
	ss_hlp = sane_open (strbuf, &hand);
	syslog (LOG_DEBUG, "sane_open: %d.\n", ss_hlp);
	sanei_printf (peer_out, "%d %d %*c\n",
		      &command, &ss_hlp, sizeof (hand), &hand);
      }
      break;

    case CMD_SANE_CLOS:
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%*c\n", sizeof (hand), &hand);
	sane_close (hand);
	syslog (LOG_DEBUG, "sane_close.\n");
	sanei_printf (peer_out, "%d\n", &command);
      }
      break;

    case CMD_SANE_GOPT:
      {
	SANE_Handle hand;
	SANE_Int optnum;
	const SANE_Option_Descriptor *od;
	const SANE_String_Const *strlist;
	const SANE_Word *wrdlist;
	const SANE_Range *rng;
	int hlp;

	sanei_scanf (peer_in, "%*c %d\n", sizeof (hand), &hand, &optnum);
	od = sane_get_option_descriptor (hand, optnum);
	if (od == NULL)
	  {
	    sanei_printf (peer_out, "%d\n", &command);
	    break;
	  }
	syslog (LOG_DEBUG, "sane_gopt: %s.\n", od->name);
	sanei_printf (peer_out, "%d %s %s %s  %d %d %d %d  %d", &command,
		      od->name, od->title, od->desc,
		      &od->type, &od->unit, &od->size, &od->cap,
		      &od->constraint_type);
	switch (od->constraint_type)
	  {
	  case SANE_CONSTRAINT_NONE:
	    break;

	  case SANE_CONSTRAINT_STRING_LIST:
	    strlist = od->constraint.string_list;
	    while (*strlist)
	      sanei_printf (peer_out, "%s", *strlist++);
	    break;

	  case SANE_CONSTRAINT_WORD_LIST:
	    wrdlist = od->constraint.word_list;
	    hlp = *wrdlist;
	    sanei_printf (peer_out, "%d", wrdlist++);
	    while (hlp--)
	      sanei_printf (peer_out, "%d", wrdlist++);
	    break;

	  case SANE_CONSTRAINT_RANGE:
	    rng = od->constraint.range;
	    sanei_printf (peer_out, "%d %d %d",
			  &rng->min, &rng->max, &rng->quant);
	    break;
	  }
	sanei_printf (peer_out, "\n");
      }
      break;

    case CMD_SANE_COPT:
      {
	SANE_Handle hand;
	SANE_Int optnum, action, type, info;
	sanei_scanf (peer_in, "%*c %d %d %d",
		     sizeof (hand), &hand, &optnum, &action, &type);
	switch (action)
	  {
	  case SANE_ACTION_SET_VALUE:
	    switch (type)
	      {
	      case SANE_TYPE_BOOL:
	      case SANE_TYPE_INT:
	      case SANE_TYPE_FIXED:
		sanei_scanf (peer_in, "%d", strbuf);
		break;

	      case SANE_TYPE_STRING:
	      case SANE_TYPE_GROUP:
		sanei_scanf (peer_in, "%s", strbuf);
		break;
	      }
	    break;
	  }
	sanei_scanf (peer_in, "\n");

	ss_hlp = sane_control_option (hand, optnum, action, strbuf, &info);
	syslog (LOG_DEBUG, "sane_copt.\n");
	sanei_printf (peer_out, "%d %d %d %d",
		      &command, &ss_hlp, &info, &type);
	switch (type)
	  {
	  case SANE_TYPE_BOOL:
	  case SANE_TYPE_INT:
	  case SANE_TYPE_FIXED:
	    sanei_printf (peer_out, "%d", strbuf);
	    break;

	  case SANE_TYPE_STRING:
	  case SANE_TYPE_GROUP:
	    sanei_printf (peer_out, "%s", strbuf);
	    break;
	  }
	sanei_printf (peer_out, "\n");
      }
      break;

    case CMD_SANE_PARM:
      {
	SANE_Handle hand;
	SANE_Parameters parm;
	sanei_scanf (peer_in, "%*c\n", sizeof (hand), &hand);
	ss_hlp = sane_get_parameters (hand, &parm);
	syslog (LOG_DEBUG, "sane_getpar.\n");
	sanei_printf (peer_out, "%d %d  %d %d %d %d %d %d\n",
		      &command, &ss_hlp,
		      &parm.format, &parm.last_frame, &parm.bytes_per_line,
		      &parm.pixels_per_line, &parm.lines, &parm.depth);
      }
      break;

    case CMD_SANE_STAR:
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%*c\n", sizeof (hand), &hand);
	ss_hlp = sane_start (hand);
	syslog (LOG_DEBUG, "sane_start.\n");
	sanei_printf (peer_out, "%d %d\n", &command, &ss_hlp);
      }
      break;

    case CMD_SANE_READ:
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%*c %d\n", sizeof (hand), &hand, &si_hlp);
	if (sizeof (strbuf) < si_hlp)
	  si_hlp = sizeof (strbuf);
	ss_hlp = sane_read (hand, strbuf, si_hlp, &si_hlp);
	syslog (LOG_DEBUG, "sane_read.\n");
	sanei_printf (peer_out, "%d %d %d %*c\n",
		      &command, &ss_hlp, &si_hlp, si_hlp, strbuf);
      }
      break;

    case CMD_SANE_ABRT:	/* Fixme ... this is not good here */
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%*c\n", sizeof (hand), &hand);
	sane_cancel (hand);
	syslog (LOG_DEBUG, "sane_cancel.\n");
      }
      break;

    case CMD_SANE_IOMO:	/* Fixme ... Emulation required ? */
      {
	SANE_Handle hand;
	sanei_scanf (peer_in, "%*c %d\n", sizeof (hand), &hand, &si_hlp);
	ss_hlp = sane_set_io_mode (hand, si_hlp);
	syslog (LOG_DEBUG, "sane_cancel.\n");
	sanei_printf (peer_out, "%d %d\n", &command, &ss_hlp);
      }
      break;
    }
}

void
say_helo (void)
{
  int cmd;
  char buffer[1024];
  char hostname[128];

  if (gethostname (hostname, sizeof (hostname)))
    *hostname = '\0';
  sprintf (buffer, "Welcome at %s. saned v0.1 at your service.", hostname);
  cmd = CMD_HELO;
  sanei_printf (peer_out, "%d %s\n", &cmd, buffer);
}

static void
quit (int signum)
{
  time_t now;

  time (&now);
  syslog (LOG_INFO, "exiting on %s\n", ctime (&now));

  closelog ();

  exit (EXIT_SUCCESS);		/* This is a nowait-daemon. */
}

int
main (int argc, char *argv[])
{
  time_t now;

  signal (SIGALRM, quit);
  signal (SIGPIPE, quit);

  peer_in = 0;
  peer_out = 0;

  openlog ("saned", LOG_PID | LOG_CONS, LOG_DAEMON);

  time (&now);
  syslog (LOG_INFO, "started on %s\n", ctime (&now));

  /* First say "Hello" to each other ... */
  say_helo ();

  /* From now on we simply respond to client requests. */
  alarm (300);
  while (sane_state != SANE_QUIT)
    readcmd ();

  quit (0);
  return EXIT_SUCCESS;
}
