/* xcam -- X-based camera frontend
   Uses the SANE library.
   Copyright (C) 1997 David Mosberger and Tristan Tarrant

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>

#include <gtk.h>
#include <gtkglue.h>
#include <sane.h>

#define NELEMS(a)	((int) sizeof(a)/sizeof(a[0]))

#define MAX_LUM		64	/* how many graylevels for 8 bit displays */

typedef struct Canvas
  {
    GtkWidget *preview;
    GdkGC *gc;
    GdkImage *gdk_image;
    GdkColormap *graylevel_cmap;	/* for 8 bit displays */
    guint32 graylevel[MAX_LUM];		/* graylevel pixels */
    GdkColormap *cube_cmap;
    GdkColor cube_colors[5 * 6 * 5];
  }
Canvas;

static const char *prog_name;
static const SANE_Device *default_device;
static GSGDialog *dialog;

static struct
  {
    GtkWidget *shell;
    GtkWidget *dialog_box;
    GtkWidget *menubar;
    GtkWidget *play_stop_label;
    GtkWidget *device_info_label;
    struct
      {
	GtkWidget *item;	/* the menu bar item */
	GtkWidget *menu;	/* the associated menu */
      }
    devices;
    Canvas canvas;
    gint gdk_input_tag;		/* tag returned by gdk_input_add () */
    int playing;		/* are we playing video? */

    SANE_Byte buf[32768];
    SANE_Int remaining;
    SANE_Parameters params;
    gpointer data;	/* image data */
    int x;		/* x position */
  }
win;

/* forward declarations: */
static void rescan_devices (GtkWidget *widget, gpointer client_data,
			    gpointer call_data);
static void next_frame (void);


#define CANVAS_EVENT_MASK	GDK_BUTTON1_MOTION_MASK |	\
				GDK_EXPOSURE_MASK |		\
				GDK_BUTTON_PRESS_MASK |		\
				GDK_ENTER_NOTIFY_MASK

static void
display_image (Canvas *canvas)
{
  if (canvas->gdk_image)
    {
      gdk_draw_image (canvas->preview->window, canvas->gc, canvas->gdk_image,
		      0, 0, 0, 0,
		      canvas->gdk_image->width, canvas->gdk_image->height);
      gdk_flush ();
    }
}

static gint
canvas_events (GtkWidget *widget, GdkEvent *event)
{
  Canvas *canvas = (Canvas *) gtk_widget_get_user_data (widget);
  if (!canvas)
    return FALSE;

  switch (event->type)
    {
    case GDK_EXPOSE:
      if (!canvas->gc)
	canvas->gc = gdk_gc_new (canvas->preview->window);
      display_image (canvas);
      break;

    case GDK_BUTTON_PRESS:
      break;

    case GDK_BUTTON_RELEASE:
      break;

    case GDK_MOTION_NOTIFY:
      break;

    case GDK_ENTER_NOTIFY:
#if 0
      gdk_colors_store (win.canvas.cube_cmap, win.canvas.cube_colors,
			NELEMS(win.canvas.cube_colors));
#endif
      break;

    default:
      break;
    }
  return FALSE;
}

static void
stop_camera (void)
{
  if (dialog)
    sane_cancel (gsg_dialog_get_device (dialog));
  if (win.gdk_input_tag >= 0)
    gdk_input_remove (win.gdk_input_tag);
  else
    win.playing = FALSE;
  win.gdk_input_tag = -1;
  if (!win.playing)
    gtk_label_set (win.play_stop_label, "Play");
}

static void
switch_device (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  SANE_Device *dev = client_data;
  char buf[512];

  if (win.playing)
    {
      win.playing = FALSE;
      stop_camera ();
    }

  if (dialog)
    gsg_destroy_dialog (dialog);

  dialog = gsg_create_dialog (win.dialog_box, dev->name, 0, 0);

  sprintf (buf, "%s %s %s", dev->vendor, dev->model, dev->type);
  gtk_label_set (win.device_info_label, buf);
}

void
device_name_dialog_cancel (GtkWidget *widget,
			   gpointer client_data, gpointer call_data)
{
  gtk_widget_destroy (client_data);
}

void
device_name_dialog_ok (GtkWidget *widget,
		       gpointer client_data, gpointer call_data)
{
  SANE_Device dev_info;
  GtkWidget *text = client_data;

  dev_info.name = gtk_text_entry_get_text (text);
  dev_info.model = "Unknown";
  dev_info.model = "";
  dev_info.type  = "";
  if (!dev_info.name)
    return;				/* huh? how come? */
  switch_device (0, &dev_info, 0);

  gtk_widget_destroy (gtk_widget_get_toplevel (text));
}

static void
prompt_for_device_name (GtkWidget *widget,
			gpointer client_data, gpointer call_data)
{
  GtkWidget *vbox, *hbox, *label, *text;
  GtkWidget *button, *dialog;

  dialog = gtk_window_new ("Device name", GTK_WINDOW_TOPLEVEL);

  /* create the main vbox */
  vbox = gtk_vbox_new (TRUE, 5);
  gtk_container_set_border_width (vbox, 5);
  gtk_container_add (dialog, vbox);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (vbox, hbox);

  label = gtk_label_new ("Device name:");
  gtk_container_add (hbox, label);
  gtk_widget_show (label);

  text = gtk_text_entry_new ();
  gtk_container_add (hbox, text);

  gtk_widget_show (hbox);

  /* the confirmation button */
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (vbox, hbox);

  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button),
		    device_name_dialog_ok, text);
  gtk_box_pack (hbox, button, TRUE, TRUE, 5, GTK_PACK_START);

  label = gtk_label_new ("OK");
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button),
		    device_name_dialog_cancel, dialog);
  gtk_box_pack (hbox, button, TRUE, TRUE, 5, GTK_PACK_START);

  label = gtk_label_new ("Cancel");
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  gtk_widget_show (hbox);
  gtk_widget_show (text);
  gtk_widget_show (vbox);
  gtk_widget_show (dialog);
}

static void
exit_callback (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  if (dialog)
    gsg_destroy_dialog (dialog);
  dialog = 0;
  exit (0);
}

static GtkWidget *
build_files_menu (void)
{
  GtkWidget *menu, *item;

  menu = gtk_menu_new ();

  item = gtk_menu_item_new ();
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  item = gtk_menu_item_new_with_label ("Exit");
  gtk_container_add (menu, item);
  gtk_callback_add (gtk_menu_item_get_state (item), exit_callback, 0);
  gtk_widget_show (item);

  return menu;
}

static GtkWidget *
build_device_menu (void)
{
  const SANE_Device ** device;
  GtkWidget *menu, *item;
  SANE_Status result;
  int i;

  menu = gtk_menu_new ();

  result = sane_get_devices (&device, SANE_FALSE);
  if (result != SANE_STATUS_GOOD)
    {
      fprintf (stderr, "%s: %s\n", prog_name, sane_strstatus (result));
      exit (1);
    }
	
  default_device = 0;
  for (i = 0; device[i]; ++i)
    {
      if (!default_device)
	default_device = device[i];

      item = gtk_menu_item_new_with_label ((char *) device[i]->name);
      gtk_container_add (menu, item);
      gtk_callback_add (gtk_menu_item_get_state (item),
			switch_device, (gpointer) device[i]);
      gtk_widget_show (item);
    }

  item = gtk_menu_item_new ();
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  item = gtk_menu_item_new_with_label ("Refresh device list...");
  gtk_callback_add (gtk_menu_item_get_state (item), rescan_devices, 0);
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  item = gtk_menu_item_new_with_label ("Specify device name...");
  gtk_callback_add (gtk_menu_item_get_state (item), prompt_for_device_name, 0);
  gtk_container_add (menu, item);
  gtk_widget_show (item);

  return menu;
}

static void
rescan_devices (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  gtk_widget_destroy (win.devices.menu);
  win.devices.menu = build_device_menu ();
  gtk_menu_item_set_submenu (win.devices.item, win.devices.menu);
}

#define READ_SANE_PIXEL(buf, buf_end, format, depth, r, g, b)		\
{									\
  switch (format)							\
    {									\
    case SANE_FRAME_GRAY:						\
      switch (depth)							\
	{								\
	case 1:								\
	  if (buf + 1 > buf_end)					\
	    goto end_of_buffer;						\
	  (r) = (g) = (b) = (*buf & src_mask) ? 0x0000 : 0xffff;	\
	  src_mask >>= 1;						\
	  if (src_mask == 0x00)						\
	    {								\
	      ++buf;							\
	      src_mask = 0x80;						\
	    }								\
	  break;							\
									\
	case 8:								\
	  if (buf + 1 > buf_end)					\
	    goto end_of_buffer;						\
	  (r) = (g) = (b) = (*buf++ << 8);				\
	  break;							\
									\
	case 16:							\
	  if (buf + 2 > buf_end)					\
	    goto end_of_buffer;						\
	  (r) = (g) = (b) = *((u_int16_t *) buf);			\
	  buf += 2;							\
	  break;							\
	}								\
      break;								\
									\
    case SANE_FRAME_RGB:						\
      switch (depth)							\
	{								\
	case 1:								\
	case 8:								\
	  if (buf + 3 > buf_end)					\
	    goto end_of_buffer;						\
	  (r) = buf[0] << 8; (g) = buf[1] << 8; (b) = buf[2] << 8;	\
	  buf += 3;							\
	  break;							\
									\
	case 16:							\
	  if (buf + 3 > buf_end)					\
	    goto end_of_buffer;						\
	  (r) = ((u_int16_t *)buf)[0];					\
	  (g) = ((u_int16_t *)buf)[1];					\
	  (b) = ((u_int16_t *)buf)[2];					\
	  buf += 6;							\
	  break;							\
	}								\
      break;								\
									\
    case SANE_FRAME_RED:						\
    case SANE_FRAME_GREEN:						\
    case SANE_FRAME_BLUE:						\
    default:								\
      fprintf (stderr, "%s: format %d not yet supported\n",		\
	       prog_name, (format));					\
    }									\
}

#define PUT_X11_PIXEL(buf, endian, depth, bpp, r, g, b, gl_map)		\
{									\
  switch (depth)							\
    {									\
    case  1: /* duh?  A Sun3 or what?? */				\
      lum = 3*(r) + 5*(g) + 2*(b);					\
      if (lum >= 5*0x8000)						\
	*buf |= dst_mask;						\
      dst_mask <<= 1;							\
      if (dst_mask > 0xff)						\
	{								\
	  buf += (bpp);							\
	  dst_mask = 0x01;						\
	}								\
      break;								\
									\
    case  8:								\
      lum = ((3*(r) + 5*(g) + 2*(b)) / (10 * 256/MAX_LUM)) >> 8;	\
      if (lum >= MAX_LUM)						\
	lum = MAX_LUM;							\
      buf[0] = (gl_map)[lum];						\
      buf += (bpp);							\
      break;								\
									\
    case 15:								\
      rgb = (  (((r) >> 11) << 10)	/* 5 bits of red */		\
	     | (((g) >> 11) <<  5)	/* 5 bits of green */		\
	     | (((b) >> 11) <<  0));	/* 5 bits of blue */		\
      ((u_int16_t *)buf)[0] = rgb;					\
      buf += (bpp);							\
      break;								\
									\
    case 16:								\
      rgb = (  (((r) >> 11) << 11)	/* 5 bits of red */		\
	     | (((g) >> 10) <<  5)	/* 6 bits of green */		\
	     | (((b) >> 11) <<  0));	/* 5 bits of blue */		\
      ((u_int16_t *)buf)[0] = rgb;					\
      buf += (bpp);							\
      break;								\
									\
    case 24:								\
      /* Is this correctly handling all byte order cases? */		\
      if ((endian) == GDK_LSB_FIRST)					\
	{								\
	  buf[0] = (b) >> 8; buf[1] = (g) >> 8; buf[2] = (r) >> 8;	\
	}								\
      else								\
	{								\
	  buf[0] = (r) >> 8; buf[1] = (g) >> 8; buf[2] = (b) >> 8;	\
	}								\
      buf += (bpp);							\
      break;								\
									\
    case 32:								\
      /* Is this correctly handling all byte order cases?  It assumes	\
	 the byte order of the host is the same as that of the		\
	 pixmap. */							\
      rgb = (((r) >> 8) << 16) | (((g) >> 8) << 8) | ((b) >> 8);	\
      ((u_int32_t *)buf)[0] = rgb;					\
      buf += (bpp);							\
      break;								\
    }									\
}

static void
input_available (gpointer ignore, gint source, GdkInputCondition cond)
{
  int x, pixels_per_line, bytes_per_line, dst_depth, src_depth;
  u_int32_t r = 0, g = 0, b = 0, lum, rgb, src_mask, dst_mask;
  size_t buf_size, remaining = win.remaining;
  SANE_Byte *src, *src_end;
  GdkByteOrder byte_order;
  u_long bytes_per_pixel;
  SANE_Frame format;
  SANE_Status result;
  SANE_Int len;
  u_char *dst;

  if (!win.playing)
    /* looks like we got cancelled */
    goto stop_and_exit;

  buf_size = sizeof (win.buf);
  format = win.params.format;
  src_depth = win.params.depth;
  dst_depth = win.canvas.gdk_image->depth;
  pixels_per_line = win.params.pixels_per_line;
  bytes_per_line = win.canvas.gdk_image->bpl;
  bytes_per_pixel = win.canvas.gdk_image->bpp;
  byte_order = win.canvas.gdk_image->byte_order;

  x = win.x;
  dst = win.data;
  src_mask = 0x80;	/* SANE has left most bit is most significant bit */
  dst_mask = 0x01;

  while (1)
    {
      result = sane_read (gsg_dialog_get_device (dialog),
			  win.buf + remaining, buf_size - remaining, &len);
      if (result != SANE_STATUS_GOOD)
	{
	  if (result == SANE_STATUS_EOF)
	    {
	      display_image (&win.canvas);
	      stop_camera ();
	      if (win.playing)
		{
		  next_frame ();		/* arrange for next frame */
		  return;
		}
	    }
	  else
	    {
	      char buf[256];
	      sprintf (buf, "Error during read: %s.", sane_strstatus (result));
	      gsg_error (buf);
	    }
	  win.playing = FALSE;
	  stop_camera ();
	  return;
	}

      src = win.buf;
      src_end = src + len + remaining;
      while (1)
	{
	  READ_SANE_PIXEL(src, src_end, format, src_depth, r, g, b);
	  PUT_X11_PIXEL(dst, byte_order, dst_depth, bytes_per_pixel, r, g, b,
			win.canvas.graylevel);
	  if (++x >= pixels_per_line)
	    {
	      x = 0;
	      dst += bytes_per_line - pixels_per_line * bytes_per_pixel;
	    }
	}
    end_of_buffer:
      remaining = src_end - src;
    }
  win.data = src;
  win.x = x;
  win.remaining = remaining;
  return;

stop_and_exit:
  win.playing = FALSE;
  stop_camera ();
  return;
}

static void
next_frame (void)
{
  char buf[256];
  SANE_Status result;
  int fd;

  result = sane_start (gsg_dialog_get_device (dialog));
  if (result != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Failed to start scanner: %s.", sane_strstatus (result));
      gsg_error (buf);
      win.playing = FALSE;
      stop_camera ();
      return;
    }

  result = sane_get_parameters (gsg_dialog_get_device (dialog), &win.params);
  if (result != SANE_STATUS_GOOD)
    {
      sprintf (buf, "Failed to get parameters: %s.", sane_strstatus (result));
      gsg_error (buf);
      win.playing = FALSE;
      stop_camera ();
      return;
    }

  if (!win.canvas.gdk_image
      || win.canvas.gdk_image->width != win.params.pixels_per_line
      || win.canvas.gdk_image->height != win.params.lines)
    {
      GdkImageType image_type = GDK_IMAGE_FASTEST;

      if (win.canvas.gdk_image)
	gdk_image_destroy (win.canvas.gdk_image);
#ifdef __alpha__
      /* Some X servers seem to have a problem with shared images that
	 have a width that is not a multiple of 8.  Duh... ;-( */
      if (win.params.pixels_per_line % 8)
	image_type = GDK_IMAGE_NORMAL;
#endif
      win.canvas.gdk_image = gdk_image_new (image_type,
					    win.canvas.preview->window->visual,
					    win.params.pixels_per_line,
					    win.params.lines);
      gtk_widget_set_usize (win.canvas.preview,
			    win.params.pixels_per_line, win.params.lines);
    }

  win.data = win.canvas.gdk_image->mem;
  win.x = 0;

  if (sane_set_io_mode (gsg_dialog_get_device (dialog), SANE_TRUE)
      == SANE_STATUS_GOOD
      && sane_get_select_fd (gsg_dialog_get_device (dialog), &fd)
      == SANE_STATUS_GOOD)
    win.gdk_input_tag = gdk_input_add (fd, GDK_INPUT_READ, input_available, 0);
  else
    input_available (0, -1, 0);
}

static void
play_stop_button (GtkWidget *widget, gpointer client_data, gpointer call_data)
{
  if (!dialog)
    return;

  if (win.playing)
    win.playing = FALSE;
  else if (win.gdk_input_tag < 0)
    {
      win.playing = TRUE;
      gtk_label_set (win.play_stop_label, "Stop");

      next_frame ();
    }
}

static void
xcam_exit (void)
{
  gtk_exit (0);
  sane_exit ();
}

int
main (int argc, char **argv)
{
  GtkWidget *menu, *menu_bar, *menu_bar_item, *label, *preview_vbox;
  GtkWidget *hbox, *vbox, *button, *alignment, *frame;
  int i;

  prog_name = strrchr (argv[0], '/');
  if (prog_name)
    ++prog_name;
  else
    prog_name = argv[0];

  sane_init (NULL, 0);

  gdk_set_debug_level (0);
  gdk_set_show_events (0);
  gtk_init (&argc, &argv);

  atexit (xcam_exit);

  win.gdk_input_tag = -1;
  win.shell = gtk_window_new ((char *) prog_name, GTK_WINDOW_TOPLEVEL);
  gtk_widget_show (win.shell);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (vbox, 0);      
  gtk_container_add (win.shell, vbox);
  gtk_widget_show (vbox);

  menu_bar = gtk_menu_bar_new ();

  win.devices.menu = build_device_menu ();

  /* "Files" entry: */
  label = gtk_label_new ("Files");
  gtk_widget_show (label);
  menu_bar_item = gtk_menu_item_new ();
  gtk_container_add (menu_bar, menu_bar_item);
  menu = build_files_menu ();
  gtk_menu_item_set_submenu (menu_bar_item, menu);
  gtk_container_add (menu_bar_item, label);
  gtk_widget_show (menu_bar_item);

  /* "Devices" entry: */
  label = gtk_label_new ("Devices");
  gtk_widget_show (label);
  win.devices.item = gtk_menu_item_new ();
  gtk_container_add (menu_bar, win.devices.item);
  gtk_container_add (win.devices.item, label);
  gtk_menu_item_set_submenu (win.devices.item, win.devices.menu);
  gtk_widget_show (win.devices.item);

  gtk_widget_show (menu_bar);

  gtk_container_add (vbox, menu_bar);

  /* add device info at top: */
  frame = gtk_frame_new (0);
  gtk_container_set_border_width (frame, 8);
  gtk_frame_set_type (frame, GTK_SHADOW_ETCHED_IN);
  gtk_container_add (vbox, frame);
  gtk_widget_show (frame);

  win.device_info_label = gtk_label_new ("");
  gtk_widget_show (win.device_info_label);
  gtk_container_add (frame, win.device_info_label);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (vbox, hbox);
  gtk_widget_show (hbox);

  /* create the device dialog box: */
  win.dialog_box = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (win.dialog_box);

  /* the preview vbox on the left hand side: */
  preview_vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack (hbox, preview_vbox, TRUE, TRUE, 0, GTK_PACK_START);

  frame = gtk_frame_new ("Preview");
  gtk_container_set_border_width (frame, 8);
  gtk_frame_set_type (frame, GTK_SHADOW_ETCHED_IN);
  gtk_container_add (preview_vbox, frame);

  alignment = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_box_pack (preview_vbox, alignment, TRUE, TRUE, 0, GTK_PACK_START);

  win.canvas.preview = gtk_drawing_area_new (320, 200, canvas_events,
					     CANVAS_EVENT_MASK);
  gtk_widget_set_user_data (win.canvas.preview, (gpointer) &win.canvas);
  gtk_container_add (frame, win.canvas.preview);

  gtk_widget_show (win.canvas.preview);
  gtk_widget_show (alignment);
  gtk_widget_show (frame);
  gtk_widget_show (preview_vbox);

  win.canvas.graylevel_cmap = gdk_colormap_get_system ();
  for (i = 0; i < NELEMS(win.canvas.graylevel); ++i)
    {
      GdkColor color;
      color.red = color.green = color.blue =
	i * 0xffff / (NELEMS(win.canvas.graylevel) - 1);
      gdk_color_alloc (win.canvas.graylevel_cmap, &color);
      win.canvas.graylevel[i] = color.pixel;
    }

#if 0
  {
      win.canvas.cube_cmap
	  = gdk_colormap_new (win.canvas.preview->window->visual, 1);
      for (i = 0; i < NELEMS(win.canvas.cube_colors); ++i)
	{
	  win.canvas.cube_colors[i].pixel = i;
	  win.canvas.cube_colors[i].red   = ((i / 30) % 5) * 0xffff / 4;
	  win.canvas.cube_colors[i].green = ((i / 5) % 6)  * 0xffff / 5;
	  win.canvas.cube_colors[i].blue  = ((i % 5))      * 0xffff / 4;
	}
      gdk_colors_store (win.canvas.cube_cmap, win.canvas.cube_colors,
			NELEMS(win.canvas.cube_colors));
      gdk_window_set_colormap (win.shell->window, win.canvas.cube_cmap);
  }
#endif

  gtk_container_add (hbox, win.dialog_box);
  if (default_device)
    switch_device (0, (gpointer) default_device, 0);

  if (dialog && gsg_dialog_get_device (dialog)
      && (sane_get_parameters (gsg_dialog_get_device (dialog), &win.params)
	  == SANE_STATUS_GOOD))
    gtk_widget_set_usize (win.canvas.preview,
			  win.params.pixels_per_line, win.params.lines);

  /* buttons at the bottom: */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), play_stop_button, 0);
  win.play_stop_label = gtk_label_new ("Play");
  gtk_container_add (button, win.play_stop_label);
  gtk_container_add (vbox, button);
  gtk_widget_show (win.play_stop_label);
  gtk_widget_show (button);

  gtk_main ();

  return 0;
}
