/* xscan -- a graphical scanner-oriented SANE frontend

  Authors:
    Andreas Beck (becka@sunserver1.rz.uni-duesseldorf.de) - Original author
    Tristan Tarrant (tristan@galactica.it) - Current maintainer
    David Mosberger-Tang (davidm@azstarnet.com)
 
  You can contact the original "The GIMP" authors at gimp@xcf.berkeley.edu
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <gtk.h>
#ifdef GIMP_PLUGIN
#include "gimp.h"
#endif

#include <gtkglue.h>
#include <sane.h>
#include <saneopts.h>
#include <progress.h>

/* Information about preview area */
struct PreviewT
  {
    gint PhysWidth, PhysHeight;
    gint PixelWidth, PixelHeight;
    double Ratio;
    gint PhysX1, PhysY1;
    gint PhysX2, PhysY2;
    gint PixelX1, PixelY1;
    gint PixelX2, PixelY2;
  }
Preview;

/* Menu description data */
typedef struct MenuItem
  {
    gchar *label;
    gchar accelerator_key;
    gint accelerator_mods;
    GtkCallback callback;
    gpointer user_data;
    struct MenuItem *subitems;
    GtkWidget *widget;
  }
MenuItem;

/* Control description */
typedef struct DlgElement
  {
    GtkWidget *widget, *label;
    GtkData *data;
    MenuItem *menu;
  }
DlgElement;

struct ScanWinT
  {
    GtkWidget *shell;
    GtkWidget *hruler;
    GtkWidget *vruler;
    GtkWidget *preview;
    GtkWidget *info_label;
    GdkGC *gc;

    GtkDataAdjustment *hruledata;	/*  horizontal data information */
    GtkDataAdjustment *vruledata;	/*  vertical data information   */
  }
ScanWin;

struct SelectionT
  {
    gint x1, y1, x2, y2;
    gint active;
  }
Selection, OldSelection;

struct DriverInfoT
  {
    gint dpi, x1, y1, x2, y2;
  }
DriverInfo;

gint LeftButtonPressed;

static GtkWidget *choose_device_dialog;
static GSGDialog *dialog;
static GdkColor RGBPalette[256];
static const SANE_Device **devlist;
static gint seldev;			/* The selected device */
static gint ndevs;			/* The number of available devices */
#ifdef GIMP_PLUGIN
static Image output;
#endif

static void 
PhysToLog (gint xx, gint yy, gint * xret, gint * yret)
{
  gfloat a, b, c, d;

  a = (Preview.PixelX2 - Preview.PixelX1) / (Preview.PhysX2 - Preview.PhysX1);
  b = Preview.PixelX1 - a * Preview.PhysX1;
  *xret = (xx - b) / a;
  c = (Preview.PixelX2 - Preview.PixelX1) / (Preview.PhysX2 - Preview.PhysX1);
  d = Preview.PixelX1 - a * Preview.PhysX1;
  *yret = (yy - d) / c;
}

static void
LogToPhys (gint xx, gint yy, gint * xret, gint * yret)
{
  gfloat a, b, c, d;

  a = (Preview.PixelX2 - Preview.PixelX1) / (Preview.PhysX2 - Preview.PhysX1);
  b = Preview.PixelX1 - a * Preview.PhysX1;
  *xret = a * xx + b;
  c = (Preview.PixelX2 - Preview.PixelX1) / (Preview.PhysX2 - Preview.PhysX1);
  d = Preview.PixelX1 - a * Preview.PhysX1;
  *yret = c * yy + d;
}

/* Create a Gtk menu widget from a MenuItem array.  */
static GtkWidget *
build_menu (MenuItem * items, GtkAcceleratorTable * table)
{
  GtkWidget *menu;
  GtkWidget *menu_item;
  GtkData *data;

  menu = gtk_menu_new ();

  while (items->label)
    {
      if (items->label[0] == '-')
	menu_item = gtk_menu_item_new ();
      else
	{
	  menu_item = gtk_menu_item_new_with_label (items->label);
	  if (items->accelerator_key && table)
	    gtk_widget_install_accelerator (menu_item,
					    table,
					    items->accelerator_key,
					    items->accelerator_mods);
	}

      gtk_container_add (menu, menu_item);

      if (items->callback)
	{
	  data = gtk_menu_item_get_state (menu_item);
	  gtk_callback_add (data, items->callback, items->user_data);
	}

      if (items->subitems)
	gtk_menu_item_set_submenu (menu_item, build_menu (items->subitems, table));

      gtk_widget_show (menu_item);
      items->widget = menu_item;

      items++;
    }

  return menu;
}

/* Update the info line with the latest size information.  */
static void 
update_param (GSGDialog *dialog, void *arg)
{
  SANE_Parameters params;
  gchar buf[200];

  if (!ScanWin.info_label)
    return;

  if (sane_get_parameters (gsg_dialog_get_device (dialog), &params)
      == SANE_STATUS_GOOD)
    {
      long size = 10*params.bytes_per_line * params.lines;
      const char *unit = "B";

      if (size >= 10*1024*1024)
	{
	  size /= 1024*1024;
	  unit = "MB";
	}
      else if (size >= 10*1024)
	{
	  size /= 1024;
	  unit = "KB";
	}
      sprintf (buf, "%dx%d: %ld.%01ld %s", params.pixels_per_line,
	       params.lines, size / 10, size % 10, unit);
    }
  else
    sprintf (buf, "Invalid parameters.");
  gtk_label_set (ScanWin.info_label, buf);
}

static void
quit_xscan (void)
{
#ifdef GIMP_PLUGIN
  if (output)
    gimp_free_image (output);
  gimp_noexit_quit ();
#endif
  if (dialog && gsg_dialog_get_device (dialog))
    sane_close (gsg_dialog_get_device (dialog));
  sane_exit ();
  gtk_exit (0);
}

/* This function is invoked whenever a cancel button is pressed. */
static void 
quit_callback (GtkWidget * widget,
	       gpointer client_data,
	       gpointer call_data)
{
  quit_xscan ();
}

/* This is invoked when the scanning is cancelled */

static void
cancel_scan_callback (GtkWidget * widget,
		      gpointer client_data, gpointer call_data)
{
  sane_cancel (gsg_dialog_get_device (dialog));
}


/* Invoked when the preview button is pressed. */

static void 
scan_preview (GtkWidget * widget,
	      gpointer client_data,
	      gpointer call_data)
{

}

/* Invoked when the zoom button is pressed */

static void 
zoom_in_preview (GtkWidget * widget,
		 gpointer client_data,
		 gpointer call_data)
{
  if (Selection.x1 >= Selection.x2
      || Selection.y1 >= Selection.y2)
    return;

  ScanWin.hruledata->lower = Selection.x1;
  ScanWin.hruledata->upper = Selection.x2;
  ScanWin.vruledata->lower = Selection.y1;
  ScanWin.vruledata->upper = Selection.y2;

  gtk_data_notify ((GtkData *) ScanWin.hruledata);
  gtk_data_notify ((GtkData *) ScanWin.vruledata);
}

static void
zoom_out_preview (GtkWidget * widget,
		  gpointer client_data,
		  gpointer call_data)
{
  ScanWin.hruledata->lower = 0;
  ScanWin.hruledata->upper = Preview.PhysWidth;
  ScanWin.vruledata->lower = 0;
  ScanWin.vruledata->upper = Preview.PhysHeight;

  gtk_data_notify ((GtkData *) ScanWin.hruledata);
  gtk_data_notify ((GtkData *) ScanWin.vruledata);
}


/* Invoked when the scan button is pressed */
static void 
scan_dialog (GtkWidget * widget,
	     gpointer client_data, gpointer call_data)
{
  SANE_Parameters parm;
  gchar *d;
  gint len, cnt, status, total;
  ProgressP scanprogress;
  FILE *fout;

#ifdef GIMP_PLUGIN
  output = NULL;
#endif
  /* Initialize device */
  status = sane_start (gsg_dialog_get_device (dialog));
  /* Is everything ok ? */
  if (status != SANE_STATUS_GOOD)
    {
      gsg_error ("Sane_start failed !");
      return;
    }
  sane_get_parameters (gsg_dialog_get_device (dialog), &parm);
#ifdef GIMP_PLUGIN
  output = gimp_new_image (0, parm.pixels_per_line, parm.lines,
	           (parm.format == SANE_FRAME_GRAY) ? GRAY_IMAGE : RGB_IMAGE);
  d = gimp_image_data (output);

#else
  fout = fopen ("sane_output.ppm", "w");
  d = (guchar *) malloc (5000);
  switch (parm.format)
    {
    case SANE_FRAME_RGB:
      fprintf (fout, "P6\n# SANE data follows\n%d %d\n255\n",
	       parm.pixels_per_line, parm.lines);
      break;

    default:
      if (parm.depth == 1)
	fprintf (fout, "P4\n# SANE data follows\n%d %d\n",
		 parm.pixels_per_line, parm.lines);
      else
	fprintf (fout, "P5\n# SANE data follows\n%d %d\n255\n",
		 parm.pixels_per_line, parm.lines);
      break;
    }

#endif
  scanprogress = progress_new ("Scanning", "Receiving data...",
			       (CancelCallback) cancel_scan_callback, d);
  gtk_main_iteration ();
  cnt = 0;
  total = parm.bytes_per_line * parm.lines;
  switch (parm.format)
    {
    case SANE_FRAME_GRAY:
    case SANE_FRAME_RGB:
      while (1)
	{
	  status = sane_read (gsg_dialog_get_device (dialog), d,
			      parm.bytes_per_line, &len);
	  if (status == SANE_STATUS_GOOD)
	    {
	      cnt += len;
#ifdef GIMP_PLUGIN
	      d += len;
#else
	      fwrite (d, 1, len, fout);
#endif
	      progress_update (scanprogress, (100 * cnt) / total);
	      gtk_main_iteration ();
	    }
	  else
	    break;
	}
      break;
    default: /* This client does not know how to handle other types of data */
      break;
    }
  progress_free (scanprogress);
  gtk_main_iteration ();
#ifndef GIMP_PLUGIN
  fclose (fout);
  free (d);
#endif

  sane_cancel (gsg_dialog_get_device (dialog));
#ifdef GIMP_PLUGIN
  if (output)
    {
      gimp_display_image (output);
      gimp_update_image (output);
      gimp_free_image (output);
    }
  gimp_noexit_quit ();
#endif
  sane_close (gsg_dialog_get_device (dialog));
  sane_exit ();
  gtk_exit (0);
}

static void 
select_device_callback (GtkWidget * widget,
			gpointer client_data, gpointer call_data)
{
  seldev = (long) client_data;
}

static void 
draw_rect (GdkWindow * win, gint x1, gint y1, gint x2, gint y2)
{
  gint xx1, yy1, width, height;

  if (x1 <= x2)
    {
      xx1 = x1;
      width = x2 - x1;
    }
  else
    {
      xx1 = x2;
      width = x1 - x2;
    }

  if (y1 <= y2)
    {
      yy1 = y1;
      height = y2 - y1;
    }
  else
    {
      yy1 = y2;
      height = y1 - y2;
    }

  gdk_draw_rectangle (win, ScanWin.gc, FALSE, xx1, yy1, width, height);
}

static void 
draw_selection (gint Redraw)
{
  gdk_gc_set_foreground (ScanWin.gc, &RGBPalette[1]);
  gdk_gc_set_background (ScanWin.gc, &RGBPalette[0]);
  gdk_gc_set_function (ScanWin.gc, GDK_INVERT);
  gdk_gc_set_line_attributes (ScanWin.gc, 1,
			      GDK_LINE_ON_OFF_DASH,
			      GDK_CAP_BUTT,
			      GDK_JOIN_MITER);
  if ((Redraw) && (OldSelection.active))
    draw_rect (ScanWin.preview->window,
	       OldSelection.x1, OldSelection.y1,
	       OldSelection.x2, OldSelection.y2);

  if (Selection.active)
    draw_rect (ScanWin.preview->window,
	       Selection.x1, Selection.y1,
	       Selection.x2, Selection.y2);
  OldSelection = Selection;
}

static SANE_Word 
set_sane_option (gint setting, gint * value)
{
  SANE_Word info, len;
  const SANE_Option_Descriptor *opt;

  if (setting < 0)
    return -1;

  opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog),
				    (long) setting);
  switch (opt->type)
    {
    case SANE_TYPE_INT:
      len = (long) *value;
      break;
    case SANE_TYPE_FIXED:
      len = ((long) *value) << SANE_FIXED_SCALE_SHIFT;
      break;
    default:
      break;
    }
  sane_control_option (gsg_dialog_get_device (dialog), (long) setting,
		       SANE_ACTION_SET_VALUE, &len, &info);
  if (info & SANE_INFO_INEXACT)
    {
      switch (opt->type)
	{
	case SANE_TYPE_INT:
	  break;
	case SANE_TYPE_FIXED:
	  len >>= SANE_FIXED_SCALE_SHIFT;
	  break;
	default:
	  break;
	}
      *value = len;
    }
  return info;
}

static void 
set_window (gint * x1, gint * y1, gint * x2, gint * y2)
{
  SANE_Word info;
  *x1 /= Preview.Ratio;
  *y1 /= Preview.Ratio;
  *x2 /= Preview.Ratio;
  *y2 /= Preview.Ratio;

  info  = set_sane_option (DriverInfo.x1, x1);
  info |= set_sane_option (DriverInfo.y1, y1);
  info |= set_sane_option (DriverInfo.x2, x2);
  info |= set_sane_option (DriverInfo.y2, y2);

  *x1 *= Preview.Ratio;
  *y1 *= Preview.Ratio;
  *x2 *= Preview.Ratio;
  *y2 *= Preview.Ratio;

  update_param (dialog, 0);
}

static void 
set_max_selection (void)
{
  const SANE_Option_Descriptor *opt;

  if (DriverInfo.x1 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog),
					DriverInfo.x1);
      Selection.x1 = SANE_UNFIX(opt->constraint.range->min) * Preview.Ratio;
    }
  if (DriverInfo.y1 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog),
					DriverInfo.y1);
      Selection.y1 = SANE_UNFIX(opt->constraint.range->min) * Preview.Ratio;
    }
  if (DriverInfo.x2 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog),
					DriverInfo.x2);
      Selection.x2 = SANE_UNFIX(opt->constraint.range->max) * Preview.Ratio;
    }
  if (DriverInfo.y2 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog),
					DriverInfo.y2);
      Selection.y2 = SANE_UNFIX(opt->constraint.range->max) * Preview.Ratio;
    }
  Selection.active = TRUE;
}

static gint 
canvas_event (GtkWidget * canvas, GdkEvent * event)
{
  gint tmp;

  switch (event->type)
    {
    case GDK_EXPOSE:
      if (!ScanWin.gc)
	ScanWin.gc = gdk_gc_new (ScanWin.preview->window);
      draw_selection (FALSE);
      break;
    case GDK_ENTER_NOTIFY:
      break;
    case GDK_LEAVE_NOTIFY:
      break;
    case GDK_BUTTON_PRESS:
      Selection.x1 = event->button.x;
      Selection.y1 = event->button.y;
      LeftButtonPressed = TRUE;
      break;
    case GDK_BUTTON_RELEASE:
      Selection.x2 = event->button.x;
      Selection.y2 = event->button.y;
      if ((Selection.x1 == Selection.x2) && (Selection.y1 == Selection.y2))
	set_max_selection ();
      else
	Selection.active = TRUE;
      if (Selection.x1 > Selection.x2)
	{
	  tmp = Selection.x1;
	  Selection.x1 = Selection.x2;
	  Selection.x2 = tmp;
	}
      if (Selection.y1 > Selection.y2)
	{
	  tmp = Selection.y1;
	  Selection.y1 = Selection.y2;
	  Selection.y2 = tmp;
	}

      set_window (&Selection.x1, &Selection.y1, &Selection.x2, &Selection.y2);
      draw_selection (TRUE);
      LeftButtonPressed = FALSE;
      break;
    case GDK_MOTION_NOTIFY:
      if (LeftButtonPressed)
	{
	  Selection.x2 = event->motion.x;
	  Selection.y2 = event->motion.y;
	  if ((Selection.x1 == Selection.x2) && (Selection.y1 == Selection.y2))
	    Selection.active = FALSE;
	  else
	    Selection.active = TRUE;
	  draw_selection (TRUE);
	}
      ScanWin.hruledata->value = (event->motion.x / Preview.Ratio);
      ScanWin.vruledata->value = (event->motion.y / Preview.Ratio);
      gtk_data_notify ((GtkData *) ScanWin.hruledata);
      gtk_data_notify ((GtkData *) ScanWin.vruledata);
      break;
    default:
      break;
    }
  return (FALSE);
}

/* Find out whether a particular capability exists */
static void 
match_option (gint * value, gint optnum, const gchar * dopt, gchar * opt)
{
  if (!dopt)
    return;
  if (!strcmp (dopt, opt))
    *value = optnum;
}

/* Fetch the device capabilities: min/max coordinates and dpi  */
static void
get_device_caps (void)
{
  gint x;
  const SANE_Option_Descriptor *opt;

  DriverInfo.dpi = -1;
  DriverInfo.x1 = -1;
  DriverInfo.y1 = -1;
  DriverInfo.x2 = -1;
  DriverInfo.y2 = -1;

  for (x = 1; (opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog), x)) != NULL; x++)
    {
      /* Find out if the options are going to be needed for the preview */
      match_option (&DriverInfo.dpi, x, opt->name, SANE_NAME_SCAN_RESOLUTION);
      match_option (&DriverInfo.x1, x, opt->name, SANE_NAME_SCAN_TL_X);
      match_option (&DriverInfo.y1, x, opt->name, SANE_NAME_SCAN_TL_Y);
      match_option (&DriverInfo.x2, x, opt->name, SANE_NAME_SCAN_BR_X);
      match_option (&DriverInfo.y2, x, opt->name, SANE_NAME_SCAN_BR_Y);
    }
}

/* Set the scan sizes */
void 
get_scan_size (void)
{
  const SANE_Option_Descriptor *opt;
  SANE_Unit unit = SANE_UNIT_MM;

  if (DriverInfo.x2 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog), DriverInfo.x2);
      unit = opt->unit;
      if (opt->type == SANE_TYPE_FIXED)
	{
	  Preview.PhysX1 = SANE_UNFIX (opt->constraint.range->min);
	  Preview.PhysX2 = SANE_UNFIX (opt->constraint.range->max);
	}
      else
	{
	  Preview.PhysX1 = opt->constraint.range->min;
	  Preview.PhysX2 = opt->constraint.range->max;
	}
      Preview.PhysWidth = Preview.PhysX2 - Preview.PhysX1;
    }
  else
    {
      Preview.PhysX1 = 0;
      Preview.PhysX2 = 100;
      Preview.PhysWidth = 100;
    }
  if (DriverInfo.y2 >= 0)
    {
      opt = sane_get_option_descriptor (gsg_dialog_get_device (dialog), DriverInfo.y2);
      unit = opt->unit;
      if (opt->type == SANE_TYPE_FIXED)
	{
	  Preview.PhysY1 = SANE_UNFIX (opt->constraint.range->min);
	  Preview.PhysY2 = SANE_UNFIX (opt->constraint.range->max);
	}
      else
	{
	  Preview.PhysY1 = opt->constraint.range->min;
	  Preview.PhysY2 = opt->constraint.range->max;
	}
      Preview.PhysHeight = Preview.PhysY2 - Preview.PhysY1;
    }
  else
    {
      Preview.PhysY1 = 0;
      Preview.PhysY2 = 100;
      Preview.PhysHeight = 100;
    }
  if (Preview.PhysHeight < 500 && unit == SANE_UNIT_PIXEL)
    {
      Preview.Ratio = 1.0;
    }
  else
    {
      Preview.Ratio = 500.0 / (float) Preview.PhysHeight;
    }
  Preview.PixelWidth = (long) (Preview.PhysWidth * Preview.Ratio);
  Preview.PixelHeight = (long) (Preview.PhysHeight * Preview.Ratio);
  Preview.PixelX1 = 0;
  Preview.PixelY1 = 0;
  Preview.PixelX2 = Preview.PixelWidth;
  Preview.PixelY2 = Preview.PixelHeight;
}

/* Create the main dialog box.  */
static void 
device_dialog (void)
{
  GtkWidget *vbox, *col1, *col2, *hbox, *button, *label, *table, *frame;
  const gchar *devname;

  devname = devlist[seldev]->name;

  /* Create the dialog box */
  ScanWin.shell = gtk_window_new ((char *) devname, GTK_WINDOW_TOPLEVEL);

  /* Create the main vbox */
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (vbox, 0);
  gtk_container_add (ScanWin.shell, vbox);
  gtk_widget_show (vbox);

  /* Top portion of window - preview and options */
  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (hbox, 3);
  gtk_box_pack (vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (hbox);

  /* Column 1 - the preview */
  col1 = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (col1, 3);
  gtk_box_pack (hbox, col1, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (col1);

  /* Column 2 - the options */
  col2 = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (col2, 3);
  gtk_box_pack (hbox, col2, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (col2);

  dialog = gsg_create_dialog (col2, devname, update_param, 0);
  if (!dialog)
    {
      gsg_error ("Could not open selected device !");
      return;
    }

  get_device_caps ();
  get_scan_size ();
  set_max_selection ();

  /* The preview */
  table = gtk_table_new (2, 2, FALSE);
  gtk_table_set_col_spacing (table, 0, 1);
  gtk_table_set_row_spacing (table, 0, 1);
  gtk_container_set_border_width (table, 2);
  gtk_container_add (col1, table);

  /* The empty box in the top-left corner */
  frame = gtk_frame_new (NULL);
  gtk_frame_set_type (frame, GTK_SHADOW_OUT);
  gtk_table_attach (table, frame, 0, 1, 0, 1, FALSE, FALSE, TRUE, 0,
		    FALSE, FALSE, TRUE, 0);
  gtk_widget_show (frame);

  /* Set the horizontal ruler */
  ScanWin.hruledata = (GtkDataAdjustment *)
    gtk_data_adjustment_new (0, Preview.PhysX1, Preview.PhysX2, 1, 1, 1);
  ScanWin.hruler = gtk_hruler_new (ScanWin.hruledata);
  gtk_ruler_set_metric (ScanWin.hruler, PIXELS);
  gtk_table_attach (table, ScanWin.hruler, 1, 2, 0, 1, TRUE, TRUE, TRUE, 0,
		    FALSE, FALSE, TRUE, 0);
  gtk_widget_show (ScanWin.hruler);

  /* Set the vertical ruler */
  ScanWin.vruledata = (GtkDataAdjustment *)
    gtk_data_adjustment_new (0, Preview.PhysY1, Preview.PhysY2, 1, 1, 1);
  ScanWin.vruler = gtk_vruler_new (ScanWin.vruledata);
  gtk_ruler_set_metric (ScanWin.vruler, PIXELS);
  gtk_table_attach (table, ScanWin.vruler, 0, 1, 1, 2, FALSE, FALSE, TRUE, 0,
		    TRUE, TRUE, TRUE, 0);
  gtk_widget_show (ScanWin.vruler);

  /* Set the preview area */
  frame = gtk_frame_new (NULL);
  gtk_frame_set_type (frame, GTK_SHADOW_IN);
  ScanWin.gc = NULL;
  ScanWin.preview =
    gtk_drawing_area_new (Preview.PixelWidth, Preview.PixelHeight,
			  canvas_event,
			  GDK_EXPOSURE_MASK | GDK_ENTER_NOTIFY_MASK |
			  GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_MASK |
			  GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_container_add (frame, ScanWin.preview);
  gtk_widget_show (ScanWin.preview);
  gtk_widget_show (frame);
  Selection.active = TRUE;
  OldSelection.active = FALSE;
  LeftButtonPressed = FALSE;

  gtk_table_attach (table, frame, 1, 2, 1, 2, TRUE, TRUE, TRUE, 0, TRUE,
		    TRUE, TRUE, 0);
  gtk_widget_show (table);

  /* The info row */
  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (hbox, 3);
  gtk_box_pack (vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (hbox);

  frame = gtk_frame_new (NULL);
  gtk_frame_set_type (frame, GTK_SHADOW_IN);
  gtk_box_pack (hbox, frame, FALSE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (frame);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (hbox, 2);
  gtk_container_add (frame, hbox);
  gtk_widget_show (hbox);

  ScanWin.info_label = gtk_label_new ("0x0: 0KB");
  gtk_label_set_alignment (ScanWin.info_label, 0.0, 0.5);
  gtk_box_pack (hbox, ScanWin.info_label, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (ScanWin.info_label);

  /* The bottom row of buttons */
  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (hbox, 3);
  gtk_box_pack (vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (hbox);

  /* The Scan button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), scan_dialog, NULL);
  label = gtk_label_new ("Scan");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  /* The Preview button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), scan_preview, NULL);
  label = gtk_label_new ("Preview");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  /* The Zoom in button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), zoom_in_preview, NULL);
  label = gtk_label_new ("Zoom in");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  /* The Zoom out button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), zoom_out_preview, NULL);
  label = gtk_label_new ("Zoom out");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  /* The Cancel button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), quit_callback, NULL);
  label = gtk_label_new ("Cancel");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  gtk_widget_show (ScanWin.shell);
}

static void 
ok_choose_dialog_callback (GtkWidget * widget,
			   gpointer client_data, gpointer call_data)
{
  gtk_widget_hide (choose_device_dialog);
  gtk_widget_destroy (choose_device_dialog);
  device_dialog ();
}

static void 
choose_device (void)
{
  GtkWidget *main_vbox, *vbox, *hbox, *button, *label;
  GtkData *owner;
  const SANE_Device *adev;
  gint i;


  choose_device_dialog = gtk_window_new ("Select device", GTK_WINDOW_TOPLEVEL);

  main_vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (main_vbox, 0);
  gtk_container_add (choose_device_dialog, main_vbox);
  gtk_widget_show (main_vbox);

  /* The list of drivers */
  vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (vbox, 3);
  gtk_box_pack (main_vbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (vbox);

  /* The radio buttons */
  owner = NULL;
  for (i = 0; i < ndevs; i++)
    {
      adev = devlist[i];

      button = gtk_radio_button_new (owner);
      gtk_callback_add (gtk_button_get_state (button),
			select_device_callback,
			(void *) (long) i);
      label = gtk_label_new ((char *) adev->name);
      gtk_label_set_alignment (label, 0.0, 0.5);
      gtk_box_pack (vbox, button, TRUE, TRUE, 0, GTK_PACK_START);
      gtk_container_add (button, label);
      gtk_widget_show (label);
      gtk_widget_show (button);
      if (i == 0)
	owner = gtk_button_get_owner (button);
    }

  /* The bottom row of buttons */
  hbox = gtk_hbox_new (FALSE, 5);
  gtk_container_set_border_width (hbox, 3);
  gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_widget_show (hbox);

  /* The Ok button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), ok_choose_dialog_callback,
		    NULL);
  label = gtk_label_new ("Ok");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  /* The Cancel button */
  button = gtk_push_button_new ();
  gtk_callback_add (gtk_button_get_state (button), quit_callback, NULL);
  label = gtk_label_new ("Cancel");
  gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
  gtk_container_add (button, label);
  gtk_widget_show (label);
  gtk_widget_show (button);

  gtk_widget_show (choose_device_dialog);
}

int 
main (int argc, char **argv)
{
  const SANE_Device *adev;
  ScanWin.info_label = NULL;
#ifdef GIMP_PLUGIN
  output = NULL;
  if (gimp_init (argc, argv))
    {
#endif
      gdk_set_debug_level (0);
      gdk_set_show_events (0);
      gtk_init (&argc, &argv);

      sane_init (0, 0);

      sane_get_devices (&devlist, SANE_FALSE);

      ndevs = 0;
      adev = devlist[0];
      while (adev)
	adev = devlist[++ndevs];
      if (ndevs > 0)
	{

	  if (ndevs == 1)
	    {
	      seldev = 0;
	      device_dialog ();
	    }
	  else
	    choose_device ();

	  gtk_main ();
	}
      else
	fprintf (stderr, "No devices available\n");

      sane_exit ();
#ifdef GIMP_PLUGIN
      gimp_noexit_quit ();
#endif
      gtk_exit (0);
#ifdef GIMP_PLUGIN
    }
  else
    fprintf (stderr, "This module can only be invoked from within the GIMP\n");
#endif
  return 0;
}
